/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "rf_common.h"

#if RF_OPT_USE_PATTERN
static rf_wchar_t get_wchar(const rf_wchar_t** ptr)
{
    return (*(*ptr)++);
}

static rf_int_t wcs_pattern_matching(   /* 0:mismatched, 1:matched */
    const rf_wchar_t* pat,    /* Matching pattern */
    const rf_wchar_t* nam,    /* String to be tested */
    rf_int_t skip,            /* Number of pre-skip chars (number of ?s) */
    rf_int_t inf              /* Infinite search (* specified) */
)
{
    const rf_wchar_t* pp, *np;
    rf_wchar_t pc, nc;
    rf_int_t nm, nx;

    while (skip--) {                /* Pre-skip name chars */
        if (!get_wchar(&nam))
            return 0; /* Branch mismatched if less name chars */
    }
    if (!*pat && inf)
        return 1;     /* (short circuit) */

    do {
        pp = pat;
        np = nam;         /* Top of pattern and name to match */
        for (;;) {
            if (*pp == '?' || *pp == '*') { /* Wildcard? */
                nm = nx = 0;
                do {                /* Analyze the wildcard chars */
                    if (*pp++ == '?') nm++;
                    else nx = 1;
                } while (*pp == '?' || *pp == '*');
                if (wcs_pattern_matching(pp, np, nm, nx))
                    return 1; /* Test new branch (recurs upto number of wildcard blocks in the pattern) */
                nc = *np;
                break;    /* Branch mismatched */
            }
            pc = get_wchar(&pp);    /* Get a pattern char */
            nc = get_wchar(&np);    /* Get a name char */
            if (pc != nc)
                break;    /* Branch mismatched? */
            if (!pc)
                return 1;      /* Branch matched? (matched at end of both strings) */
        }
        get_wchar(&nam);            /* nam++ */
    } while (inf && nc);            /* Retry until end of name if infinite search is specified */
    return 0;
}
#endif

#if RF_OPT_USE_NTFS
rf_bool_t rf_wstr_in_pattern(const rf_wchar_t* name, const rf_wchar_t* expression)
{
    if (expression == NULL)
        return TRUE;
    /* '*' ƥκļ */
    if (expression[0] == '*' && expression[1] == 0)
        return TRUE;

#if RF_OPT_USE_PATTERN
    return wcs_pattern_matching(expression, name, 0, 0);
#else
    //RF_PRINTF(RF_DBG_LEVEL_ERROR, "error: need define RF_OPT_USE_PATTERN == 1\n");
    return RF_EINVAL;
#endif
}
#endif

#if RF_OPT_USE_PATTERN
static rf_wchar_t get_achar(const rf_tchar_t** ptr)
{
    rf_wchar_t chr;

    chr = (rf_u8_t) * (*ptr)++;
    if (IsLower(chr))
        chr -= 0x20;              /* To upper ASCII char */
#if RF_OPT_HAVE_CODE_PAGE
    if (isdbcs1((rf_char_t)chr) && isdbcs2(**ptr))         /* Get DBC 2nd byte if needed */
        chr = chr << 8 | (rf_u8_t) * (*ptr)++;
    if (chr >= 0x80 && chr < 0x100)
        chr = e2A(chr);
#endif
    return chr;
}

static rf_int_t pattern_matching(   /* 0:mismatched, 1:matched */
    const rf_tchar_t* pat,    /* Matching pattern */
    const rf_tchar_t* nam,    /* String to be tested */
    rf_int_t skip,            /* Number of pre-skip chars (number of ?s) */
    rf_int_t inf              /* Infinite search (* specified) */
)
{
    const rf_tchar_t* pp, *np;
    rf_wchar_t pc, nc;
    rf_int_t nm, nx;

    while (skip--) {                /* Pre-skip name chars */
        if (!get_achar(&nam))
            return 0; /* Branch mismatched if less name chars */
    }
    if (!*pat && inf)
        return 1;     /* (short circuit) */

    do {
        pp = pat;
        np = nam;         /* Top of pattern and name to match */
        for (;;) {
            if (*pp == '?' || *pp == '*') { /* Wildcard? */
                nm = nx = 0;
                do {                /* Analyze the wildcard chars */
                    if (*pp++ == '?') nm++;
                    else nx = 1;
                } while (*pp == '?' || *pp == '*');
                if (pattern_matching(pp, np, nm, nx))
                    return 1; /* Test new branch (recurs upto number of wildcard blocks in the pattern) */
                nc = *np;
                break;    /* Branch mismatched */
            }
            pc = get_achar(&pp);    /* Get a pattern char */
            nc = get_achar(&np);    /* Get a name char */
            if (pc != nc)
                break;    /* Branch mismatched? */
            if (!pc)
                return 1;      /* Branch matched? (matched at end of both strings) */
        }
        get_achar(&nam);            /* nam++ */
    } while (inf && nc);            /* Retry until end of name if infinite search is specified */
    return 0;
}
#endif

rf_bool_t rf_str_in_pattern(const rf_char_t* name, const rf_char_t* expression)
{
    if (expression == NULL)
        return TRUE;
    /* '*' ƥκļ */
    if (expression[0] == '*' && expression[1] == 0)
        return TRUE;

#if RF_OPT_USE_PATTERN
    return pattern_matching(expression, name, 0, 0);
#else
    RF_PERROR("error: need define RF_OPT_USE_PATTERN == 1\n", 1, return RF_EINVAL);
#endif
}
