/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#define _RF_MAX_NAME 256
#include "../common/rf_common.h"
#include "ext2.h"
#include "ext2_volume.h"
#include "ext2_file.h"

static rf_int_t _file_search(rf_ext2_file_ref_t dir, const rf_char_t* name, rf_dirent_ref_t pfd);

/* ͨڵ ȡڵڵ */
rf_int_t ext2_get_inode_entry(rf_ext2_volume_ref_t vol, rf_u32_t inum, rf_fentry_ref_t entry)
{
    rf_int_t ret;
    rf_u32_t group, index, blknum;
    rf_u32_t block_offset, group_inode_table;

    if (inum == 0)
        return RF_ERROR;

    group = (inum - 1) / vol->inodes_per_group;
    if (group > vol->total_groups)
        return RF_ERROR;

    ret = rf_disk_read(vol->dev_handle, vol->group_start_sector + ((group << 5) >> vol->sector_size_bits), 1, vol->sector_buffer);
    if (ret)
        return ret;

    group_inode_table = u32_read(vol->sector_buffer, (rf_u16_t)(((group << 5) % vol->sector_size) + 8));

    index = ((inum - 1) % vol->inodes_per_group) * vol->inode_size;
    block_offset = index % vol->block_size;
    blknum = group_inode_table + (index >> vol->block_size_bits);

    entry->sector = (blknum << vol->sectors_per_cluster_bits) + (block_offset >> vol->sector_size_bits);
    entry->offset = (rf_u16_t)(block_offset % vol->sector_size);
    return 0;
}

/* ͨļƫ ȡ */
rf_int_t ext2_file_get_cluster(rf_ext2_file_ref_t file, rf_u32_t* cluster)
{
    rf_int_t ret;
    rf_lloff_t last_sector;
    rf_u32_t sector_offset;
    rf_u32_t clust;

    *cluster = 0;
    if (file->block.type == TYPE_DIRECT_BLOCK) {
        if (file->last_cache_sector != file->entry.sector) {
            ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
            if (ret)
                return ret;
            file->last_cache_sector = file->entry.sector;
        }
        if (file->block.u.direct.offset >= EXT2_NDIR_BLOCKS)
            return RF_ESUCCESS;

        *cluster = u32_read(file->vol->sector_buffer, file->entry.offset + 0x28 + (file->block.u.direct.offset << 2));
        file->block.u.direct.offset++;
        if (file->block.u.direct.offset == EXT2_NDIR_BLOCKS) {
            file->block.type = TYPE_INDIRECT_BLOCK;
            clust = u32_read(file->vol->sector_buffer, file->entry.offset + 0x58);
            file->block.u.indirect.sector = clust << file->vol->sectors_per_cluster_bits;
            file->block.u.indirect.offset = 0;
        }
    } else if (file->block.type == TYPE_INDIRECT_BLOCK) {
        if (file->block.u.indirect.sector != 0) {
            last_sector = file->block.u.indirect.sector + (file->block.u.indirect.offset >> file->vol->sector_size_bits);
            if (file->last_cache_sector != last_sector) {
                ret = rf_disk_read(file->vol->dev_handle, last_sector, 1, file->vol->sector_buffer);
                if (ret)
                    return ret;
                file->last_cache_sector = last_sector;
            }
            *cluster = u32_read(file->vol->sector_buffer, file->block.u.indirect.offset % file->vol->sector_size);
        }

        file->block.u.indirect.offset += 4;
        if (file->block.u.indirect.offset == file->vol->block_size) {
            ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
            if (ret)
                return ret;

            file->block.type = TYPE_DOUBLE_INDIRECT_BLOCK;
            clust = u32_read(file->vol->sector_buffer, file->entry.offset + 0x5c);
            file->block.u.double_indirect.indirect_sector = clust << file->vol->sectors_per_cluster_bits;
            file->block.u.double_indirect.indirect_offset = 4;

            if (file->block.u.double_indirect.indirect_sector != 0) {
                ret = rf_disk_read(file->vol->dev_handle, file->block.u.double_indirect.indirect_sector, 1, file->vol->sector_buffer);
                if (ret)
                    return ret;

                clust = u32_read(file->vol->sector_buffer, 0);
                file->block.u.double_indirect.direct_sector = clust << file->vol->sectors_per_cluster_bits;
            } else {
                file->block.u.double_indirect.direct_sector = 0;
            }
            file->block.u.double_indirect.direct_offset = 0;
        }
    } else if (file->block.type == TYPE_DOUBLE_INDIRECT_BLOCK) {
        if (file->block.u.double_indirect.direct_sector != 0) {
            last_sector = file->block.u.double_indirect.direct_sector +
                          (file->block.u.double_indirect.direct_offset >> file->vol->sector_size_bits);
            if (file->last_cache_sector != last_sector) {
                ret = rf_disk_read(file->vol->dev_handle, last_sector, 1, file->vol->sector_buffer);
                if (ret)
                    return ret;
                file->last_cache_sector = last_sector;
            }
            *cluster = u32_read(file->vol->sector_buffer, file->block.u.double_indirect.direct_offset % file->vol->sector_size);
        }

        file->block.u.double_indirect.direct_offset += 4;
        if (file->block.u.double_indirect.direct_offset == file->vol->block_size) {
            if (file->block.u.double_indirect.indirect_offset == file->vol->block_size) {
                ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
                if (ret)
                    return ret;

                file->block.type = TYPE_TRIPLE_INDIRECT_BLOCK;
                clust = u32_read(file->vol->sector_buffer, file->entry.offset + 0x60);
                file->block.u.triple_indirect.double_sector = clust << file->vol->sectors_per_cluster_bits;
                file->block.u.triple_indirect.double_offset = 4;

                if (file->block.u.triple_indirect.double_sector != 0) {
                    ret = rf_disk_read(file->vol->dev_handle, file->block.u.triple_indirect.double_sector, 1, file->vol->sector_buffer);
                    if (ret)
                        return ret;

                    clust = u32_read(file->vol->sector_buffer, 0);
                    file->block.u.triple_indirect.indirect_sector = clust << file->vol->sectors_per_cluster_bits;
                    file->block.u.triple_indirect.indirect_offset = 4;

                    if (file->block.u.triple_indirect.indirect_sector != 0) {
                        ret = rf_disk_read(file->vol->dev_handle, file->block.u.triple_indirect.indirect_sector, 1, file->vol->sector_buffer);
                        if (ret)
                            return ret;
                        clust = u32_read(file->vol->sector_buffer, 0);
                        file->block.u.triple_indirect.direct_sector = clust << file->vol->sectors_per_cluster_bits;
                    } else {
                        file->block.u.triple_indirect.direct_sector = 0;
                    }
                } else {
                    file->block.u.triple_indirect.direct_sector = 0;
                }
                file->block.u.triple_indirect.direct_offset = 0;
                return 0;
            }

            sector_offset = file->block.u.double_indirect.indirect_offset >> file->vol->sector_size_bits;
            if (file->block.u.double_indirect.indirect_sector != 0) {
                ret = rf_disk_read(file->vol->dev_handle, file->block.u.double_indirect.indirect_sector + sector_offset, 1, file->vol->sector_buffer);
                if (ret)
                    return ret;

                file->block.u.double_indirect.direct_sector = u32_read(file->vol->sector_buffer,
                        file->block.u.double_indirect.indirect_offset % file->vol->sector_size) << file->vol->sectors_per_cluster_bits;
            } else {
                file->block.u.double_indirect.direct_sector = 0;
            }

            file->block.u.double_indirect.direct_offset = 0;
            file->block.u.double_indirect.indirect_offset += 4;
        }
    } else if (file->block.type == TYPE_TRIPLE_INDIRECT_BLOCK) {
        if (file->block.u.triple_indirect.direct_sector != 0) {
            last_sector = file->block.u.triple_indirect.direct_sector +
                          (file->block.u.triple_indirect.direct_offset >> file->vol->sector_size_bits);
            if (file->last_cache_sector != last_sector) {
                ret = rf_disk_read(file->vol->dev_handle, last_sector, 1, file->vol->sector_buffer);
                if (ret)
                    return ret;
                file->last_cache_sector = last_sector;
            }
            *cluster = u32_read(file->vol->sector_buffer, file->block.u.triple_indirect.direct_offset % file->vol->sector_size);
        }

        file->block.u.triple_indirect.direct_offset += 4;
        if (file->block.u.triple_indirect.direct_offset == file->vol->block_size) {
            if (file->block.u.triple_indirect.indirect_offset == file->vol->block_size) {
                if (file->block.u.triple_indirect.double_offset == file->vol->block_size)
                    return RF_ERROR;

                file->block.u.triple_indirect.indirect_offset = 0;
                if (file->block.u.triple_indirect.double_sector != 0) {
                    sector_offset = file->block.u.triple_indirect.double_offset >> file->vol->sector_size_bits;
                    ret = rf_disk_read(file->vol->dev_handle, file->block.u.triple_indirect.double_sector + sector_offset, 1, file->vol->sector_buffer);
                    if (ret)
                        return ret;

                    clust = u32_read(file->vol->sector_buffer, file->block.u.triple_indirect.double_offset % file->vol->sector_size);
                    file->block.u.triple_indirect.indirect_sector = clust << file->vol->sectors_per_cluster_bits;
                } else {
                    file->block.u.triple_indirect.indirect_sector = 0;
                }
                file->block.u.triple_indirect.double_offset += 4;
            }

            file->block.u.triple_indirect.direct_offset = 0;
            if (file->block.u.triple_indirect.indirect_sector != 0) {
                sector_offset = file->block.u.triple_indirect.indirect_offset >> file->vol->sector_size_bits;
                ret = rf_disk_read(file->vol->dev_handle, file->block.u.triple_indirect.indirect_sector + sector_offset, 1, file->vol->sector_buffer);
                if (ret)
                    return ret;
                clust = u32_read(file->vol->sector_buffer, file->block.u.triple_indirect.indirect_offset % file->vol->sector_size);
                file->block.u.triple_indirect.direct_sector = clust << file->vol->sectors_per_cluster_bits;
            } else {
                file->block.u.triple_indirect.direct_sector = 0;
            }
            file->block.u.triple_indirect.indirect_offset += 4;
        }
    }
    return 0;
}

rf_int_t ext2_file_gather_clusters(rf_ext2_file_ref_t file)
{
    rf_u8_t index = 0;
    rf_u32_t cluster, next_cluster;
    rf_int_t ret;

    file->gather_cluester_bit = FALSE;
    chain_init(&file->clust, TRUE);

    cluster = file->last_cluster;
    file->clust.chain[0].cluster = file->last_cluster;
    file->clust.chain[0].number = 1;
    file->blocks_offset++;
    if (file->blocks_offset << file->vol->block_size_bits >= file->length) {
        file->gather_cluester_bit = TRUE;
        return RF_ESUCCESS;
    }

    /* 󻺴 Ч */
    file->last_cache_sector = 0;
    for (;;) {
        ret = ext2_file_get_cluster(file, &next_cluster);
        if (ret)
            return ret;

        if ((next_cluster == (cluster + 1)) && (file->clust.chain[index].number != UCHAR_MAX)) {
            file->clust.chain[index].number++;
        } else {
            index++;
            if (index == RF_MAX_CHAIN_LENGTH) {
                file->last_cluster = next_cluster;
                file->clust.last_chain_index = index - 1;
                return RF_ESUCCESS;
            }
            file->clust.chain[index].number = 1;
            file->clust.chain[index].cluster = next_cluster;
        }

        file->blocks_offset++;
        if (file->blocks_offset << file->vol->block_size_bits >= file->length) {
            file->gather_cluester_bit = TRUE;
            file->clust.last_chain_index = index;
            return RF_ESUCCESS;
        }
        cluster = next_cluster;
    }
    return 0;
}

static rf_u32_t mode2attr(rf_u16_t mode)
{
    rf_u32_t attr = 0;
    if (Ext2IsReadOnly(mode))
        attr |= RF_ATTR_READONLY;
    if (S_ISDIR(mode))
        attr |= RF_ATTR_DIRECTORY;
    return attr;
}

static rf_u32_t file2attr(rf_u8_t type)
{
    /*
    http://www.nongnu.org/ext2-doc/ext2.html
    EXT2_FT_UNKNOWN     0   Unknown File Type
    EXT2_FT_REG_FILE    1   Regular File
    EXT2_FT_DIR         2   Directory File
    EXT2_FT_CHRDEV      3   Character Device
    EXT2_FT_BLKDEV      4   Block Device
    EXT2_FT_FIFO        5   Buffer File
    EXT2_FT_SOCK        6   Socket File
    EXT2_FT_SYMLINK     7   Symbolic Link
    */
    if (type == 2)
        return RF_ATTR_DIRECTORY;
    /* Ӧøʵ  */
    return RF_ATTR_NORMAL;
}

rf_int_t ext2_file_raw_open(rf_ext2_file_ref_t file, rf_fentry_t* fentry, rf_u32_t open_flag)
{
    rf_int_t ret;
    rf_u16_t i_mode;

    file->sector_buf_flag = FALSE;
    if (open_flag & RF_O_SECBUF)
        file->sector_buf_flag = TRUE;

    ret = rf_disk_read(file->vol->dev_handle, fentry->sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;

    file->entry.sector = fentry->sector;
    file->entry.offset = fentry->offset;

    file->length = (rf_u64_t)u32_read(file->vol->sector_buffer, fentry->offset + 4) |
                   ((rf_u64_t)u32_read(file->vol->sector_buffer, fentry->offset + 0x6c) << 32);

    i_mode = u16_read(file->vol->sector_buffer, fentry->offset);
    file->attr = mode2attr(i_mode);

    file->blocks_offset = 0;
    file->block.type = TYPE_DIRECT_BLOCK;
    file->block.u.direct.offset = 0;
    file->last_cache_sector = 0;
    file->current_offset = 0;

    chain_init(&file->clust, TRUE);
    file->clust.chain_offset = 1;
    file->gather_cluester_bit = FALSE;
    if (file->length) {
        ret = ext2_file_get_cluster(file, &file->last_cluster);
        if (RF_ESUCCESS != ret)
            return ret;
        if (file->sector_buf_flag) {
            ret = ext2_file_gather_clusters(file);
            if (RF_ESUCCESS != ret)
                return ret;
        }
    } else {
        file->gather_cluester_bit = TRUE;
    }
    return 0;
}

static rf_int_t _file_open(rf_ext2_file_ref_t file, const rf_char_t* name, rf_u32_t open_flag)
{
    rf_int_t ret;
    rf_dirent_t fdir;

    fdir.name = NULL;
    fdir.len = 0;
    ret = _file_search(file, name, &fdir);
    if (ret != RF_ESUCCESS)
        return ret;
    file->entry.sector = fdir.entry.sector;
    file->entry.offset = fdir.entry.offset;
    return ext2_file_raw_open(file, &file->entry, open_flag);
}

rf_int_t ext2_file_open(rf_ext2_file_ref_t file, const rf_char_t* name, rf_int_t flags)
{
    rf_int_t ret;
    rf_int_t last_path = 0;
    const rf_tchar_t* fn;
    rf_fentry_t fentry;
    rf_char_t* buf;
    DEF_PATH_BUF

    fn = name;
    file->access_flag = flags & 0x3;
    /*  ֧· 򿪷ǰĿ¼ */
    if (is_absolute_path(name)) {
        fentry = file->vol->root_dir_entry;
    } else {
        fentry = file->vol->current_dir_entry;
    }
    ret = ext2_file_raw_open(file, &fentry, RF_O_RDONLY);
    if (ret != RF_ESUCCESS)
        return ret;

    GET_PATH_BUF(buf);
    do {
        ret = path_get_fn(&fn, buf, RF_MAX_FNAME, file->vol->code_page);
        if (ret != RF_ESUCCESS)
            break;
        ret = path_is_last(&fn, &last_path);
        if (ret != RF_ESUCCESS)
            break;
        /* ȡļ */
        if (buf[0] == '\0')
            break;
        /* 򿪵ǵǰĿ¼  */
        if (buf[0] ==  '.' && buf[1] == '\0')
            continue;

        if (last_path == 1) {
            ret = _file_open(file, buf, flags);
        } else { /* мΪĿ¼ ʹֻ  */
            ret = _file_open(file, buf, RF_O_RDONLY);
        }

        if (ret != RF_ESUCCESS) {
            /* һĿ¼ ļĿ¼ */
            if (ret == RF_ENDFILE && last_path == 1 && flags & RF_O_CREAT)
                ret = RF_EPERM;
            break;
        }

        /* ֻһ ļ  */
        RF_PERROR("Path to the include file error\n", last_path == 0 && !(file->attr & RF_ATTR_DIRECTORY), ret = RF_ENOTDIR;);

    } while (last_path == 0);
    FREE_PATH_BUF(buf);
    if (ret)
        return ret;

    if (file->attr & RF_ATTR_DIRECTORY)
        return RF_ESUCCESS;
    return  RF_ESUCCESS;
}

static rf_int_t clusters_file_read(rf_ext2_file_ref_t file, rf_u32_t read_sector, rf_u32_t* pread_sector, rf_u8_t* buf)
{
    rf_u32_t tmp, count;
    rf_int_t ret;

    *pread_sector = 0;
    while (read_sector) {
        if (file->clust.chain_offset > file->clust.last_chain_index) {
            if (file->gather_cluester_bit == FALSE) {
                ret = ext2_file_gather_clusters(file);
                /* һ  ֱӷ */
                if (ret == RF_ENDFILE)
                    return RF_ESUCCESS;
                if (ret != RF_ESUCCESS)
                    return ret;
            } else {
                /* ļݶ ֱӷ */
                return RF_ESUCCESS;
            }
        }

        /* count_secΪ32λ ֹDSP numberΪ16BITʱ ˷ */
        /* ʣ */
        tmp = ((file->clust.chain[file->clust.chain_offset].number - file->clust.cluster_offset) <<
               file->vol->sectors_per_cluster_bits) - file->clust.sector_offset;
        if (tmp >= read_sector) {
            count = read_sector;
            read_sector = 0;
        } else {
            count = tmp;
            read_sector -= tmp;
        }

        if (buf) {
            if (file->clust.chain[file->clust.chain_offset].cluster != 0) {
                ret = rf_disk_read(file->vol->dev_handle,
                                   (((file->clust.chain[file->clust.chain_offset].cluster + file->clust.cluster_offset) <<
                                     file->vol->sectors_per_cluster_bits) + file->clust.sector_offset),
                                   count, buf + ((*pread_sector) << file->vol->sector_size_bits));
                if (ret)
                    return ret;
            } else {
                /* ϡļ */
                memset(buf + ((*pread_sector) << file->vol->sector_size_bits), 0, count << file->vol->sector_size_bits);
            }
        }
        *pread_sector += count;
        chain_offset_sector(&file->clust, file->vol->sectors_per_cluster, count);
    }
    return RF_ESUCCESS;
}

rf_int_t ext2_file_read(rf_ext2_file_ref_t file, rf_void_t* buf, rf_u32_t len, rf_u32_t* plen)
{
    rf_int_t ret;
    rf_u32_t read, tmp;

    if (file->access_flag == RF_O_WRONLY)
        return RF_EACCES;
    if (file->sector_buf_flag)
        RF_PERROR("error: it is necessary to read the whole sectors\n", len != file->vol->sector_size, return RF_EINVAL);

    *plen = 0;
    /* Ŀ¼ļ ûļȴС м ؽΪĿ¼ */
    if (!(file->attr & RF_ATTR_DIRECTORY)) {
        if ((file->current_offset + len) > file->length)
            len = (rf_u32_t)(file->length - file->current_offset);
    }
    if (len == 0)
        return RF_ESUCCESS;

    if (file->clust.byte_offset != 0) {
_READ_LAST:
        if (buf)
            ret = clusters_file_read(file, 1, &read, file->vol->sector_buffer);
        else
            ret = clusters_file_read(file, 1, &read, NULL);
        if (ret != RF_ESUCCESS)
            return ret;

        if (read == 1) {
            tmp = min((rf_u32_t)(file->vol->sector_size - file->clust.byte_offset), len);
            /* buf=pff->vol->sector_buffer ڶȡĿ¼ Ҫ */
            if (buf && buf != file->vol->sector_buffer)
                memcpy((rf_u8_t*)buf + *plen, file->vol->sector_buffer + file->clust.byte_offset, tmp);
            len -= tmp;
            file->clust.byte_offset += (rf_u16_t)tmp;
            if (file->clust.byte_offset % file->vol->sector_size == 0)
                file->clust.byte_offset = 0;
            *plen += tmp;
            file->current_offset += tmp;
        } else {
            /* ļôض ļȽС ֱӷ */
            return RF_ESUCCESS;
        }

        if (len == 0) {
            if (file->clust.byte_offset != 0)
                chain_back_offset_sector(&file->clust, file->vol->sectors_per_cluster, 1);
            return RF_ESUCCESS;
        }
    }

    read = len >> file->vol->sector_size_bits;
    if (read) {
        if (buf)
            ret = clusters_file_read(file, read, &tmp, (rf_u8_t*)buf + *plen);
        else
            ret = clusters_file_read(file, read, &tmp, NULL);
        if (ret != RF_ESUCCESS)
            return ret;

        len -= tmp << file->vol->sector_size_bits;
        *plen += tmp << file->vol->sector_size_bits;
        file->current_offset += tmp << file->vol->sector_size_bits;
        /* ȡСҪʱļ */
        if (tmp < read)
            return RF_ESUCCESS;
    }

    /* ȡβһ */
    len %= file->vol->sector_size;
    if (len)
        goto _READ_LAST;
    return RF_ESUCCESS;
}

rf_int_t ext2_file_close(rf_ext2_file_ref_t file)
{
    file->open_flag = FALSE;
    return RF_ESUCCESS;
}

rf_int_t ext2_file_get_stat(rf_ext2_file_ref_t file, rf_stat_ref_t pstat)
{
    pstat->entry.sector = file->entry.sector;
    pstat->entry.offset = file->entry.offset;
    pstat->st_mode = file->attr;
    pstat->st_size = file->length;
    return RF_ESUCCESS;
}

rf_int_t ext2_file_get_ftime(rf_ext2_file_ref_t file, rf_file_ftime_ref_t pt)
{
    rf_int_t ret;

    ret = rf_disk_read(file->vol->dev_handle, file->entry.sector, 1, file->vol->sector_buffer);
    if (ret)
        return ret;

    pt->create.high_date_time = 0;
    pt->create.low_date_time = u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + 0x0c);
    pt->last_access.high_date_time = 0;
    pt->last_access.low_date_time = u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + 0x8);
    pt->last_write.high_date_time = 0;
    pt->last_write.low_date_time = u32_read(file->vol->sector_buffer, (rf_u16_t)file->entry.offset + 0x10);
    return RF_ESUCCESS;
}

static rf_int_t _file_search(rf_ext2_file_ref_t dir, const rf_char_t* name, rf_dirent_ref_t pfd)
{
    rf_int_t ret;
    rf_u32_t len, tmp;
    rf_u32_t inode, offset;
    rf_u16_t rec_len;
    rf_u8_t name_len, file_type;
    rf_bool_t match_flag, wildcard;
    rf_char_t* buf;
    DEF_FIND_BUF

    if (!(dir->attr & RF_ATTR_DIRECTORY))
        return RF_EINVAL;
    /* ȡʱ Ŀ¼ƫ */
    offset = (rf_u32_t)(dir->current_offset) % dir->vol->sector_size;
    dir->current_offset -= offset;

    /* Ŀ¼ݵ з */
    ret = ext2_file_read(dir, dir->vol->sector_buffer, dir->vol->sector_size, &len);
    if (ret != RF_ESUCCESS)
        return ret;
    if (!len)
        return RF_ENDFILE;

    GET_FIND_BUF(buf);
    wildcard = rf_str_contain_wild_cards(name);
    do {
        inode = u32_read(dir->vol->sector_buffer, offset);
        offset += 4;
        if (offset == dir->vol->sector_size) {
            ret = ext2_file_read(dir, dir->vol->sector_buffer, dir->vol->sector_size, &len);
            if (ret != RF_ESUCCESS)
                break;
            if (!len) {
                ret = RF_ENDFILE;
                break;
            }
            offset = 0;
        }
        rec_len = u16_read(dir->vol->sector_buffer, offset);
        offset += 2;
        name_len = u8_read(dir->vol->sector_buffer, offset);
        offset++;
        file_type = u8_read(dir->vol->sector_buffer, offset);
        offset++;
        if (offset == dir->vol->sector_size) {
            ret = ext2_file_read(dir, dir->vol->sector_buffer, dir->vol->sector_size, &len);
            if (ret != RF_ESUCCESS)
                break;
            if (!len) {
                ret = RF_ENDFILE;
                break;
            }
            offset = 0;
        }

        tmp = min(name_len, dir->vol->sector_size - offset);
        if (inode != 0)
            memcpy(buf, dir->vol->sector_buffer + offset, tmp);
        offset += tmp;
        if (tmp < name_len) {
            /* ļԽ */
            ret = ext2_file_read(dir, dir->vol->sector_buffer, dir->vol->sector_size, &len);
            if (ret != RF_ESUCCESS)
                break;
            if (!len) {
                ret = RF_ENDFILE;
                break;
            }
            offset = name_len - tmp;
            if (inode != 0)
                memcpy(buf + tmp, dir->vol->sector_buffer, offset);
        }
        buf[name_len] = 0;
        /* ļƫ4ֽڶ */
        offset = ((offset + 3) >> 2) << 2;
        tmp = rec_len - (8 + (((name_len + 3) >> 2) << 2));
        offset += tmp;

        match_flag = TRUE;
        /* Чмڵ */
        if (inode == 0)
            match_flag = FALSE;
        if (!wildcard) {
            /* ƥַ ֱӱȽ */
            if (strcmp(buf, name))
                match_flag = FALSE;
        } else {
            /* ļƥַǷƥ */
            if (!rf_str_in_pattern(buf, name))
                match_flag = FALSE;
        }

        if (!match_flag) {
            if (offset < dir->vol->sector_size)
                continue;

            /*  ƫƵN */
            len = (offset - dir->vol->sector_size) >> dir->vol->sector_size_bits;
            if (len) {
                ret = ext2_file_read(dir, NULL, len << dir->vol->sector_size_bits, &tmp);
                if (ret != RF_ESUCCESS)
                    break;
                if (!len) {
                    ret = RF_ENDFILE;
                    break;
                }
            }

            /*  һ */
            offset -= len << dir->vol->sector_size_bits;
            offset %= dir->vol->sector_size;
            tmp = 1;
            ret = ext2_file_read(dir, dir->vol->sector_buffer, dir->vol->sector_size, &tmp);
            if (ret != RF_ESUCCESS)
                break;
            if (!tmp) {
                ret = RF_ENDFILE;
                break;
            }
            continue;
        }

        if (pfd) {
            ret = ext2_get_inode_entry(dir->vol, inode, &pfd->entry);
            if (ret != RF_ESUCCESS)
                break;
            pfd->attribute = file2attr(file_type);
            pfd->type = dir->vol->code_page;
            if (pfd->name) {
                if (pfd->len < name_len + 1) {
                    ret = RF_EBUFSMALL;
                    break;
                }
                strcpy(pfd->name, buf);
            }

            /* ƫڵǰΧ */
            if (offset < dir->vol->sector_size) {
                dir->current_offset -= dir->vol->sector_size;
                dir->current_offset += offset;
                ret = chain_back_offset_sector(&dir->clust, dir->vol->sectors_per_cluster, 1);
                break;
            }

            /* мN */
            len = (offset - dir->vol->sector_size) >> dir->vol->sector_size_bits;
            if (len) {
                ret = ext2_file_read(dir, NULL, len << dir->vol->sector_size_bits, &tmp);
                if (ret != RF_ESUCCESS)
                    break;
                offset %= dir->vol->sector_size;
                dir->current_offset += offset;
            }
        }
        ret = RF_ESUCCESS;
        break;
    } while (1);
    FREE_FIND_BUF(buf);
    return ret;
}

rf_int_t ext2_file_read_dir(rf_ext2_file_ref_t file, const rf_char_t* pattern, rf_dirent_ref_t pfd)
{
    return _file_search(file, pattern, pfd);
}

rf_int_t ext2_file_lseek(rf_ext2_file_ref_t file, rf_u64_t offset, rf_int_t origin)
{
    return RF_EINVAL;
}
