/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "fat_volume.h"
#include "fat_file.h"
#include "fat.h"

/* յĴ(FAT12/16/32) */
#define EMPTY_CLUSTER       ((rf_u32_t)0X00000000)
/* ڵڴ˴ΪһFAT12 */
#define LAST_CLUSTER_12     ((rf_u32_t)0X00000FF7)
/* ڵڴ˴ΪһFAT16 */
#define LAST_CLUSTER_16     ((rf_u32_t)0X0000FFF7)
/* ڵڴ˴Ϊһ(FAT32) */
#define LAST_CLUSTER_32     ((rf_u32_t)0X0FFFFFF7)
/* ڸFAT */
#define UPDATE_CLUSTER      ((rf_u32_t)0X0FFFFFFF)

static rf_int_t find_bitmap( /* 0:Not found, 2..:Cluster block found, 0xFFFFFFFF:Disk error */
    rf_fat_volume_t* fs,    /* Filesystem object */
    rf_u32_t clst,  /* Cluster number to scan from */
    rf_u32_t ncl,   /* Number of contiguous clusters to find (1..) */
    rf_u32_t* rcl
);

static rf_int_t change_bitmap(
    rf_fat_volume_t* fs,    /* Filesystem object */
    rf_u32_t clst,  /* Cluster number to change from */
    rf_u32_t ncl,   /* Number of clusters to be changed */
    rf_int_t bv     /* bit value to be set (0 or 1) */
);

#if RF_OPT_FS_READONLY == 0
static rf_int_t sync_window(rf_fat_volume_t* vol)
{
    rf_int_t ret;
    rf_u32_t wsect;
    rf_int_t nf;

    if (vol->write_flag) {
        wsect = vol->current_work_sector;
        ret = rf_disk_write(vol->dev_handle, wsect, 1, vol->sector_buffer);
        if (ret)
            return ret;
        vol->write_flag = 0;
        if (wsect - vol->bpb_fat_start_sector < vol->bpb_fat_size) {
            for (nf = vol->bpb_num_fats; nf >= 2; nf--) {
                wsect += vol->bpb_fat_size;
                ret = rf_disk_write(vol->dev_handle, wsect, 1, vol->sector_buffer);
                if (ret)
                    return ret;
            }
        }
    }
    return 0;
}
#endif

static rf_int_t move_window(rf_fat_volume_t* vol, rf_u32_t sector)
{
    rf_int_t ret;

    if (sector != vol->current_work_sector) {
#if RF_OPT_FS_READONLY == 0
        if (sync_window(vol) != 0)
            return RF_EIO;
#endif

        ret = rf_disk_read(vol->dev_handle, sector, 1, vol->sector_buffer);
        if (ret)
            return ret;
        vol->current_work_sector = sector;
    }
    return RF_ESUCCESS;
}

rf_int_t read_clusters(rf_fat_file_ref_t file, rf_u32_t clst, rf_u32_t* next_clst)
{
    rf_u32_t wc, bc;
    rf_u8_t* p;
    rf_int_t ret;

    RF_PERROR("clst error\n", clst < 2 || clst >= file->vol->bpb_total_clusters + 2, return RF_ERROR);

    switch (file->vol->fs_type) {
#if RF_OPT_USE_FAT12
    case RF_TYPE_FAT12 :
        bc = (rf_u32_t)clst;
        bc += bc >> 1;
        ret = move_window(file->vol, file->vol->bpb_fat_start_sector + (bc >> file->vol->sector_size_bits));
        if (ret != 0)
            return ret;
        wc = file->vol->sector_buffer[bc % file->vol->sector_size];
        bc++;
        ret = move_window(file->vol, file->vol->bpb_fat_start_sector + (bc >> file->vol->sector_size_bits));
        if (ret != 0)
            return ret;
        wc |= file->vol->sector_buffer[bc % file->vol->sector_size] << 8;
        *next_clst = clst & 1 ? wc >> 4 : (wc & 0xFFF);
        if (*next_clst > LAST_CLUSTER_12)
            return 0;
        break;
#endif

#if RF_OPT_USE_FAT16
    case RF_TYPE_FAT16 :
        ret = move_window(file->vol, file->vol->bpb_fat_start_sector + (clst >> (file->vol->sector_size_bits - 1)));
        if (ret != 0)
            return ret;
        p = &file->vol->sector_buffer[(clst << 1) % file->vol->sector_size ];
        *next_clst =  u16_read(p, 0);
        if (*next_clst > LAST_CLUSTER_16)
            return 0;
        break;
#endif

#if RF_OPT_USE_FAT32
    case RF_TYPE_FAT32 :
        ret = move_window(file->vol, file->vol->bpb_fat_start_sector + (clst >> (file->vol->sector_size_bits - 2)));
        if (ret != 0)
            return ret;
        p = &file->vol->sector_buffer[(clst << 2) % file->vol->sector_size];
        *next_clst = u32_read(p, 0) & 0x0FFFFFFF;
        if (*next_clst > LAST_CLUSTER_32)
            return 0;
        break;
#endif

#if RF_OPT_USE_EXFAT
    case RF_TYPE_EXFAT:
        if (file->objsize != 0) {
            rf_u32_t cofs = clst - file->start_cluster; /* Offset from start cluster */
            rf_u32_t clen = (rf_u32_t)((file->objsize - 1) >> file->vol->sector_size_bits) / file->vol->bpb_total_clusters; /* Number of clusters - 1 */

            if (file->stat == 2 && cofs <= clen) {  /* Is it a contiguous chain? */
                *next_clst = (cofs == clen) ? 0x7FFFFFFF : clst + 1;    /* No data on the FAT, generate the value */
                break;
            }
            if (file->stat == 3 && cofs < file->n_cont) {   /* Is it in the 1st fragment? */
                *next_clst = clst + 1;  /* Generate the value */
                break;
            }
            if (file->stat != 2) {  /* Get value from FAT if FAT chain is valid */
                if (file->n_frag != 0) {    /* Is it on the growing edge? */
                    *next_clst = 0x7FFFFFFF;    /* Generate EOC */
                } else {
                    ret = move_window(file->vol, file->vol->bpb_fat_start_sector + (clst >> (file->vol->sector_size_bits - 2)));
                    if (ret != 0)
                        return ret;
                    p = &file->vol->sector_buffer[(clst << 2) % file->vol->sector_size];
                    *next_clst = u32_read(p, 0) & 0x0FFFFFFF;
                }
                break;
            }
        }
        /* go to default */
#endif

    default:
        RF_PERROR("Invalid file system type\n", 1, return RF_ERROR);
    }
    RF_PERROR("clst error\n", (*next_clst >= file->vol->bpb_total_clusters + 2 || *next_clst == clst), return RF_ERROR);
    return 0;
}

/* Ѽļص */
rf_int_t clusters_gather_file(rf_fat_file_ref_t file, rf_bool_t first)
{
    rf_u32_t cluster, next_cluster = 0;
    rf_u16_t index = 0;
    rf_int_t ret;

    if (first == TRUE) {
        cluster = file->start_cluster;
    } else {
        cluster = file->clust.chain[file->clust.last_chain_index].cluster
                  + file->clust.chain[file->clust.last_chain_index].number - 1;
    }

    file->gather_all_clust_bit = FALSE;
    chain_init(&file->clust, TRUE);

    if (cluster == 0) {
        /* ļжϻ Ϊļ */
        file->clust.chain_offset = 1;
        file->gather_all_clust_bit = TRUE;
        return RF_ESUCCESS;
    }

    file->clust.chain[0].cluster = cluster;
    file->clust.chain[0].number = 1;
    file->vol->current_work_sector = 0xffffffff;
    do {
        ret = read_clusters(file, cluster, &next_cluster);
        if (ret != RF_ESUCCESS)
            return ret;
        RF_PERROR("clst error\n", next_cluster < 2, return RF_ERROR);

        switch (file->vol->fs_type) {
#if RF_OPT_USE_FAT32
        case RF_TYPE_FAT32: {
            /* Ӧýһȷ0xfff7ж!!! */
            if (next_cluster > LAST_CLUSTER_32)
                file->gather_all_clust_bit = TRUE;
        }
        break;
#endif

#if RF_OPT_USE_FAT16
        case RF_TYPE_FAT16: {
            if (next_cluster > LAST_CLUSTER_16)
                file->gather_all_clust_bit = TRUE;
        }
        break;
#endif

#if RF_OPT_USE_FAT12
        case RF_TYPE_FAT12: {
            if (next_cluster > LAST_CLUSTER_12)
                file->gather_all_clust_bit = TRUE;
        }
        break;
#endif
        }

        if (!first) {
            first = TRUE;
            if (file->gather_all_clust_bit == TRUE)
                return RF_ENDFILE;
            file->clust.chain[0].cluster = next_cluster;
            cluster = next_cluster;
            continue;
        }

        if ((next_cluster == (cluster + 1)) && (file->clust.chain[index].number != UCHAR_MAX)) {
            file->clust.chain[index].number++;
        } else {
            index++;
            /* ļһ, */
            /* ػȡ ļݻûռ */
            if (file->gather_all_clust_bit == TRUE || index == RF_MAX_CHAIN_LENGTH) {
                file->clust.last_chain_index = index - 1;
                return RF_ESUCCESS;
            }
            file->clust.chain[index].number = 1;
            file->clust.chain[index].cluster = next_cluster;
        }
        cluster = next_cluster;
    } while (1);
    return RF_ERROR;
}

rf_int_t clusters_offset_sector(rf_fat_file_ref_t file, rf_u32_t count)
{
    if (file->root_dir_flag && (file->vol->fs_type == RF_TYPE_FAT12 || file->vol->fs_type == RF_TYPE_FAT16)) {
        if (file->size + count > file->vol->bpb_root_sector_size)
            return RF_ENDFILE;
        file->size += count;
        return RF_ESUCCESS;
    }
    return chain_offset_sector(&file->clust, file->vol->sectors_per_cluster, count);
}

rf_int_t clusters_back_offset_sector(rf_fat_file_ref_t file, rf_u32_t count)
{
    if (file->root_dir_flag && (file->vol->fs_type == RF_TYPE_FAT12 || file->vol->fs_type == RF_TYPE_FAT16)) {
        if (count > file->size)
            return RF_ERROR;
        file->size -= count;
        return RF_ESUCCESS;
    }
    return chain_back_offset_sector(&file->clust, file->vol->sectors_per_cluster, count);
}

rf_int_t clusters_file_read(rf_fat_file_ref_t file, rf_u32_t read_sector, rf_u32_t* pread_sector, rf_u8_t* buf)
{
    rf_u32_t tmp, count;
    rf_int_t ret;

    *pread_sector = 0;
    if (file->root_dir_flag && (file->vol->fs_type == RF_TYPE_FAT12 || file->vol->fs_type == RF_TYPE_FAT16)) {
        /* Ŀ¼ļ length Ϊ ȡƫ Լһ */
        *pread_sector = min(file->vol->bpb_root_sector_size - (rf_u32_t)file->size, read_sector);
        if (!*pread_sector)
            return RF_ENDFILE;
        if (buf) {
            file->vol->current_work_sector = file->vol->bpb_root_start_sector + (rf_u32_t)file->size;
            ret = rf_disk_read(file->vol->dev_handle, file->vol->current_work_sector, 1, buf);
            if (ret)
                return ret;
        }
        file->size += *pread_sector;
        return RF_ESUCCESS;
    }

    /* fat1216 ǸĿ¼ļ fat32 Ŀ¼ļ */
    while (read_sector) {
        if (file->clust.chain_offset > file->clust.last_chain_index) {
            if (file->gather_all_clust_bit == FALSE) {
                ret = clusters_gather_file(file, file->first_gather_cluester_bit);
                file->first_gather_cluester_bit = FALSE;
                /* һ  ֱӷ */
                if (ret == RF_ENDFILE)
                    return RF_ESUCCESS;
                if (ret != RF_ESUCCESS)
                    return ret;
            } else {
                /* ļݶ ֱӷ */
                return RF_ESUCCESS;
            }
        }

        /*count_secΪ32λ ֹDSP numberΪ16BITʱ ˷*/
        /*ʣ*/
        tmp = ((file->clust.chain[file->clust.chain_offset].number -
                file->clust.cluster_offset) << file->vol->sectors_per_cluster_bits) - file->clust.sector_offset;
        if (tmp >= read_sector) {
            count = read_sector;
            read_sector = 0;
        } else {
            count = tmp;
            read_sector -= tmp;
        }

        if (buf) {
            file->vol->current_work_sector = (file->vol->fat_data_start_sector +
                                              (file->clust.chain[file->clust.chain_offset].cluster + file->clust.cluster_offset - 2) *
                                              file->vol->sectors_per_cluster + file->clust.sector_offset);
            ret = rf_disk_read(file->vol->dev_handle, file->vol->current_work_sector,
                               count, buf + ((*pread_sector) << file->vol->sector_size_bits));
            if (ret)
                return ret;
        }
        *pread_sector += count;
        clusters_offset_sector(file, count);
    }

#if RF_OPT_FILE_SECBUF
    if (file->sector_buf_flag && file->clust.chain_offset > file->clust.last_chain_index && file->gather_all_clust_bit == FALSE) {
        ret = clusters_gather_file(file, FALSE);
        if (ret == RF_ENDFILE)
            return RF_ESUCCESS;
        if (ret != RF_ESUCCESS)
            return ret;
    }
#endif
    return RF_ESUCCESS;
}

/* Ѽմص */
rf_int_t clusters_gather_empty(rf_fat_file_ref_t file, rf_u32_t number_clst, rf_u32_ref_t pnumber_clst, rf_bool_t gather)
{
    rf_u32_t start_cluster, cluster, next_cluster;
    rf_int_t ret;
#if RF_OPT_USE_EXFAT
    rf_u32_t count_cluster = 0;
#endif

    *pnumber_clst = 0;
    cluster = file->vol->fat_fsi_nxt_free;
    if (!cluster || cluster >= file->vol->bpb_total_clusters + 2)
        cluster = 2;
    start_cluster = cluster;

    chain_init(&file->clust, TRUE);
    while (number_clst) {
#if RF_OPT_USE_EXFAT
        if (file->vol->fs_type == RF_TYPE_EXFAT) {
            ret = find_bitmap(file->vol, start_cluster, 1, &count_cluster);
            if (ret)
                return ret;
            if (count_cluster == 0 || count_cluster == 0xFFFFFFFF)
                return RF_ENOSPC;
            ret = change_bitmap(file->vol, count_cluster, 1, 1);
            if (ret)
                return ret;
        } else
#endif
        {
            ret = read_clusters(file, cluster, &next_cluster);
        }
        if (ret != RF_ESUCCESS)
            return ret;

        if (next_cluster == 0) {
            (*pnumber_clst)++;
            number_clst--;

            /* մ */
            if (gather) {
                if (file->clust.chain[0].cluster == 0) {
                    file->clust.chain[0].cluster = cluster;
                    file->clust.chain[0].number = 1;
                } else {
                    if (((file->clust.chain[file->clust.last_chain_index].cluster +
                            file->clust.chain[file->clust.last_chain_index].number) == cluster) &&
                            (file->clust.chain[file->clust.last_chain_index].number != UCHAR_MAX)) {
                        file->clust.chain[file->clust.last_chain_index].number++;
                    } else {
                        file->clust.last_chain_index++;
                        if (file->clust.last_chain_index == RF_MAX_CHAIN_LENGTH) {
                            file->clust.last_chain_index = RF_MAX_CHAIN_LENGTH - 1;
                            break;
                        }
                        file->clust.chain[file->clust.last_chain_index].number = 1;
                        file->clust.chain[file->clust.last_chain_index].cluster = cluster;
                    }
                }
            }
        }

        cluster++;
        /* β ӳ */
        if (cluster >= file->vol->bpb_total_clusters + 2)
            cluster = 2;
        if (cluster == start_cluster)
            break;
    }

    if (gather)
        file->vol->fat_fsi_nxt_free = cluster;
    return RF_ESUCCESS;
}

#if RF_OPT_FS_READONLY == 0
rf_int_t write_clusters(rf_fat_file_ref_t file, rf_u32_t clst, rf_u32_t val)
{
    rf_u32_t bc;
    rf_u8_t* p;
    rf_int_t res;

    RF_PERROR("clst error\n", (clst < 2 || clst >= file->vol->bpb_total_clusters + 2), return RF_ERROR);
    switch (file->vol->fs_type) {
#if RF_OPT_USE_FAT12
    case RF_TYPE_FAT12 :
        bc = (rf_u32_t)clst;
        bc += bc >> 1;
        res = move_window(file->vol, file->vol->bpb_fat_start_sector + (bc >> file->vol->sector_size_bits));
        if (res != 0)
            break;
        p = &file->vol->sector_buffer[bc % file->vol->sector_size];
        *p = (clst & 1) ? ((*p & 0x0F) | ((rf_u8_t)val << 4)) : (rf_u8_t)val;
        bc++;
        file->vol->write_flag = 1;
        res = move_window(file->vol, file->vol->bpb_fat_start_sector + (bc >> file->vol->sector_size_bits));
        if (res != 0)
            break;
        p = &file->vol->sector_buffer[bc % file->vol->sector_size];
        *p = (clst & 1) ? (rf_u8_t)(val >> 4) : ((*p & 0xF0) | ((rf_u8_t)(val >> 8) & 0x0F));
        break;
#endif

#if RF_OPT_USE_FAT16
    case RF_TYPE_FAT16:
        res = move_window(file->vol, file->vol->bpb_fat_start_sector + (clst >> (file->vol->sector_size_bits - 1)));
        if (res != 0)
            break;
        p = &file->vol->sector_buffer[(clst << 1) % file->vol->sector_size];
        u16_write(p, 0, (rf_u16_t)val);
        break;
#endif

#if RF_OPT_USE_FAT32 ||  RF_OPT_USE_EXFAT
    case RF_TYPE_FAT32:
    case RF_TYPE_EXFAT:
        res = move_window(file->vol, file->vol->bpb_fat_start_sector + (clst >> (file->vol->sector_size_bits - 2)));
        if (res != 0)
            break;
        p = &file->vol->sector_buffer[(clst << 2) % file->vol->sector_size];
        u32_write(p, 0, val);
        break;
#endif

    default:
        return RF_ERROR;
    }
    file->vol->write_flag = 1;
    return res;
}

#if RF_OPT_USE_EXFAT && !RF_OPT_FS_READONLY
static rf_int_t find_bitmap( /* 0:Not found, 2..:Cluster block found, 0xFFFFFFFF:Disk error */
    rf_fat_volume_t* fs,    /* Filesystem object */
    rf_u32_t clst,  /* Cluster number to scan from */
    rf_u32_t ncl,   /* Number of contiguous clusters to find (1..) */
    rf_u32_t* rcl
)
{
    rf_u8_t bm, bv;
    rf_u32_t i;
    rf_u32_t val, scl, ctr;
    rf_int_t ret;

    clst -= 2;  /* The first bit in the bitmap corresponds to cluster #2 */
    if (clst >= fs->bpb_total_clusters - 2)
        clst = 0;
    scl = val = clst;
    ctr = 0;
    for (;;) {
        ret = move_window(fs, fs->bpb_fat_start_sector + val / 8 / fs->sector_size);
        if (ret)
            return ret;
        i = val / 8 % fs->sector_size;
        bm = 1 << (val % 8);
        do {
            do {
                bv = fs->sector_buffer[i] & bm;
                bm <<= 1;       /* Get bit value */
                if (++val >= fs->bpb_total_clusters - 2) {  /* Next cluster (with wrap-around) */
                    val = 0;
                    bm = 0;
                    i = fs->sector_size;
                }
                if (!bv) {  /* Is it a free cluster? */
                    if (++ctr == ncl) {
                        *rcl = scl + 2; /* Check if run length is sufficient for required */
                        return 0;
                    }
                } else {
                    scl = val;
                    ctr = 0;     /* Encountered a cluster in-use, restart to scan */
                }
                if (val == clst) {
                    *rcl = 0;
                    return 0;   /* All cluster scanned? */
                }
            } while (bm);
            bm = 1;
        } while (++i < fs->sector_size);
    }
}

static rf_int_t change_bitmap(
    rf_fat_volume_t* fs,    /* Filesystem object */
    rf_u32_t clst,  /* Cluster number to change from */
    rf_u32_t ncl,   /* Number of clusters to be changed */
    rf_int_t bv     /* bit value to be set (0 or 1) */
)
{
    rf_u8_t bm;
    rf_u32_t i;
    rf_u32_t sect;
    rf_int_t ret;


    clst -= 2;  /* The first bit corresponds to cluster #2 */
    sect = fs->bpb_fat_start_sector + clst / 8 / fs->sector_size;   /* Sector address (assuming bitmap is located top of the cluster heap) */
    i = clst / 8 % fs->sector_size;  /* Byte offset in the sector */
    bm = 1 << (clst % 8);   /* Bit mask in the byte */
    for (;;) {
        ret = move_window(fs, sect++);
        if (ret)
            return ret;
        do {
            do {
                if (bv == (int)((fs->sector_buffer[i] & bm) != 0))
                    return RF_ERROR; /* Is the bit expected value? */
                fs->sector_buffer[i] ^= bm; /* Flip the bit */
                fs->write_flag = 1;
                if (--ncl == 0)
                    return 0;   /* All bits processed? */
            } while (bm <<= 1);     /* Next bit */
            bm = 1;
        } while (++i < fs->sector_size);        /* Next byte */
        i = 0;
    }
}
#endif

rf_int_t _clusters_write(rf_fat_file_ref_t file, rf_u32_t begin_cluster,
                         rf_u32_t number, rf_u32_t* last_cluster, rf_bool_t mode)
{
    rf_int_t ret;
    rf_u32_t clst;

    clst = begin_cluster;
    if (mode == F_DESIGN_FAT && *last_cluster != 0) {
        ret = write_clusters(file, *last_cluster, clst);
        if (ret)
            return ret;
        /* ڽļĵط Ϊ */
        if (number == 1 && clst == 0xffffffff)
            goto _EXIT;
    }

    while (number) {
        if (number > 1) {
            if (mode == F_DESIGN_FAT)
                ret = write_clusters(file, clst, clst + 1);
            else
                ret = write_clusters(file, clst, EMPTY_CLUSTER);
        } else {
            /* һ Ҫر */
            if (mode == F_DESIGN_FAT)
                ret = write_clusters(file, clst, 0xffffffff);
            else {
                ret = read_clusters(file, clst, last_cluster);
                if (ret)
                    break;
                ret = write_clusters(file, clst, EMPTY_CLUSTER);
            }
        }
        if (ret)
            return ret;

        clst++;
        number--;
    }

_EXIT:
#if RF_OPT_USE_EXFAT
    if (file->vol->fs_type == RF_TYPE_EXFAT) {
        ret = change_bitmap(file->vol, begin_cluster, number, mode);
        if (ret != 0)
            return 0;
    }
#endif
    return 0;
}

/* ļ￪ʼɾ 0ʾɾļд */
rf_int_t clusters_file_delete(rf_fat_file_ref_t file)
{
    rf_int_t ret = RF_ESUCCESS;
    rf_u32_t tmp_clst, last_clst, min_clst, count_clst = 0;
    rf_u32_t offset;

    min_clst = file->vol->fat_fsi_nxt_free;
    tmp_clst = file->start_cluster;
    file->vol->current_work_sector = 0xffffffff;
    /* ǰλ ж!!! */
    if (file->current_offset != 0) {
        /* ļβ ܲü */
        if (file->clust.chain_offset > file->clust.last_chain_index)
            return RF_ESUCCESS;

        /* ɾǰļ ʣ */
        min_clst = min(file->clust.chain[file->clust.chain_offset].cluster + file->clust.cluster_offset + 1, min_clst);
        last_clst = 0;
        count_clst += file->clust.chain[file->clust.chain_offset].number - file->clust.cluster_offset;
        ret = _clusters_write(file, file->clust.chain[file->clust.chain_offset].cluster +
                              file->clust.cluster_offset,
                              file->clust.chain[file->clust.chain_offset].number - file->clust.cluster_offset,
                              &last_clst,
                              F_DELETE_FAT);
        if (ret != RF_ESUCCESS)
            return ret;

        /* ļβ Ϊ */
        offset = file->clust.chain[file->clust.chain_offset].cluster + file->clust.cluster_offset - 1;
        ret = _clusters_write(file, 0xffffffff, 1, &offset, F_DESIGN_FAT);
        if (ret != RF_ESUCCESS)
            return ret;
        sync_window(file->vol);
        offset = ++file->clust.chain_offset;
    } else {
        tmp_clst = 0;
        offset = 0;
    }

    do {
        while (offset <= file->clust.last_chain_index) {
            min_clst = min(file->clust.chain[offset].cluster, min_clst);
            count_clst += file->clust.chain[offset].number;
            ret = _clusters_write(file, file->clust.chain[offset].cluster, file->clust.chain[offset].number, &last_clst, F_DELETE_FAT);
            if (ret != RF_ESUCCESS)
                return ret;
            offset++;
        }
        sync_window(file->vol);

        if (file->gather_all_clust_bit)
            break;

        offset = 0;
        file->start_cluster = last_clst;
        ret = clusters_gather_file(file, TRUE);
        if (RF_ESUCCESS != ret)
            break;
    } while (1);
    if (ret)
        return ret;

    /* ɾļдҪ ³ʼر */
    if (file->current_offset == 0) {
        file->clust.chain[0].cluster = 0;
        file->clust.chain_offset = 1;
        file->clust.last_chain_index = 0;
        file->gather_all_clust_bit = TRUE;
    }
    file->start_cluster = tmp_clst;
    if (RF_ESUCCESS != ret)
        return ret;

    file->vol->update_fsi_flag = TRUE;
    file->vol->fat_fsi_nxt_free = min_clst;
    if (file->vol->fat_fsi_free_count != 0xffffffff)
        file->vol->fat_fsi_free_count += count_clst;
    return RF_ESUCCESS;
}

/* ļβ count_sector */
rf_int_t clusters_file_build_end(rf_fat_file_ref_t file, rf_u32_t count_sector, rf_u32_t* pcount_sector)
{
    rf_int_t ret;
    rf_u32_t i;
    rf_u32_t clst, count_clst, begin_clst, last_clst, last_clst2, gather_clst;

    begin_clst = 0;
    count_clst = 0;
    *pcount_sector = 0;
    file->vol->current_work_sector = 0xffffffff;
    if (file->clust.chain[0].cluster != 0) {
        last_clst2 = last_clst = file->clust.chain[file->clust.last_chain_index].cluster +
                                 file->clust.chain[file->clust.last_chain_index].number - 1;
    } else {
        last_clst = 0;
    }

    clst = (count_sector + file->vol->sectors_per_cluster - 1) / file->vol->sectors_per_cluster;
    while (clst) {
        ret = clusters_gather_empty(file, clst, &gather_clst, TRUE);
        if (ret)
            return ret;
        if (!gather_clst)
            break;

        if (begin_clst == 0)
            begin_clst = file->clust.chain[0].cluster;

        for (i = 0; i <= file->clust.last_chain_index; i++) {
            count_clst += file->clust.chain[i].number;
            ret = _clusters_write(file, file->clust.chain[i].cluster,
                                  file->clust.chain[i].number, &last_clst, F_DESIGN_FAT);
            if (ret != RF_ESUCCESS)
                return ret;
            last_clst = file->clust.chain[i].cluster + file->clust.chain[i].number - 1;
        }

        *pcount_sector += (gather_clst << file->vol->sectors_per_cluster_bits);
        clst -= gather_clst;
    }
    /* пռ ֱӷ*/
    if (!count_clst)
        return RF_ESUCCESS;

    ret = sync_window(file->vol);
    if (ret)
        return ret;

    file->vol->update_fsi_flag = TRUE;
    if (file->vol->fat_fsi_free_count != 0xffffffff)
        file->vol->fat_fsi_free_count -= count_clst;
    if (file->start_cluster == 0) {
        file->attrib_change_bit = TRUE;
        file->start_cluster = begin_clst;
        ret = clusters_gather_file(file, TRUE);
    } else {
        file->clust.last_chain_index = 0;
        file->clust.chain[0].cluster = last_clst2;
        file->clust.chain[0].number = 1;
        ret = clusters_gather_file(file, FALSE);
    }
    return ret;
}

rf_int_t clusters_file_write_sector(rf_fat_file_ref_t file, rf_u8_t* buf, rf_u32_t count_sector, rf_u32_t* pcount_sector)
{
    rf_int_t ret;
    rf_u32_t count, tmp;
    rf_lloff_t sector_offset;

    *pcount_sector = 0;
    while (count_sector) {
        /* Ѽ  ûпʼѼ */
        if (file->clust.chain_offset > file->clust.last_chain_index) {
            /* дѼ չռ */
            if (file->gather_all_clust_bit) {
                ret = clusters_file_build_end(file, count_sector, &tmp);
                if (ret)
                    return ret;
                if (!tmp)
                    return RF_ENOSPC;
            } else {
                ret = clusters_gather_file(file, FALSE);
                if (ret)
                    return ret;
            }
        }

        count = (file->clust.chain[file->clust.chain_offset].number - file->clust.cluster_offset) *
                file->vol->sectors_per_cluster - file->clust.sector_offset;
        count = min(count, count_sector);
        if (buf) {
            sector_offset = file->vol->fat_data_start_sector + ((file->clust.chain[file->clust.chain_offset].cluster +
                            file->clust.cluster_offset - 2) << file->vol->sectors_per_cluster_bits) + file->clust.sector_offset;

            RF_PERROR("sector_offset error\n", sector_offset < file->vol->fat_data_start_sector, return RF_ERROR);

            ret = rf_disk_write(file->vol->dev_handle, sector_offset,
                                count, buf + ((*pcount_sector) << file->vol->sector_size_bits));
            if (ret)
                return ret;
        }

        *pcount_sector += count;
        clusters_offset_sector(file, count);
        count_sector -= count;
    }
    /*  βռ */
    if (file->sector_buf_flag && file->clust.chain_offset > file->clust.last_chain_index && file->gather_all_clust_bit) {
        /* βһ  ļд */
        ret = clusters_file_build_end(file, 1, &tmp);
        if (ret)
            return ret;
        if (!tmp)
            return RF_ENOSPC;
    }
    return RF_ESUCCESS;
}
#endif
