/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "fat_volume.h"
#include "fat_file.h"
#include "fat.h"
#include "fat_mkfs.h"
#include "../src/rf_fs.h"

#define _RF_FAT_MODULE
#define RF_COMM_MAX_VOLUME             RF_OPT_FAT_MAX_VOLUME
#define RF_COMM_MAX_FILE               RF_OPT_FAT_MAX_FILE
#define RF_COMM_INIT_FS_TYPE           RF_TYPE_FAT32
#define rf_volume_comm_t               rf_fat_volume_t
#define rf_volume_comm_ref_t           rf_fat_volume_ref_t
#define rf_file_comm_t                 rf_fat_file_t
#define rf_file_comm_ref_t             rf_fat_file_ref_t
#define comm_get_volume_handle_count   fat_get_volume_handle_count
#define comm_get_file_handle_count     fat_get_file_handle_count
#define comm_check_file_access         fat_check_file_access
#define init_data                      fat_init
#define vol_alloc                      fat_vol_alloc
#define vol_free                       fat_vol_free
#include "../common/rf_comdata.h"

#if RF_OPT_USE_MKFS
rf_int_t fat_mkfs(const rf_char_t* devname, const rf_void_t* ctx)
{
    rf_fat_volume_ref_t vol;
    rf_int_t ret;

    vol = fat_vol_alloc();
    if (vol == NULL)
        return RF_EMFILE;
    ret = fat_mkfs_imp(devname, vol, (rf_fat_mkfs_param_t*)ctx);
    fat_vol_free(vol);
    return ret;
}
#endif

rf_int_t fat_mount(struct rf_vol_ctx* vctx, rf_mount_param_t* param)
{
    rf_fat_volume_ref_t vol;
    rf_int_t ret;

    vol = fat_vol_alloc();
    if (vol == NULL)
        return RF_EMFILE;

    ret = fat_volume_open(vol, param);
    if (ret == RF_ESUCCESS) {
        vctx->vol = vol;
        vol->open_flag = 1;
        return RF_ESUCCESS;
    }
    if (vol->dev_handle != NULL)
        rf_disk_close(vol->dev_handle);
#if RF_OPT_USE_MALLOC
    if (vol->sector_buffer != NULL)
        rf_free(vol->sector_buffer);
#endif
    fat_vol_free(vol);
    return ret;
}

rf_int_t fat_unmount(struct rf_vol_ctx* vctx)
{
    rf_int_t ret;

    ret = fat_volume_close((rf_fat_volume_ref_t)vctx->vol);
    if (ret == RF_ESUCCESS)
        fat_vol_free(vctx->vol);
    return ret;
}

#if RF_OPT_USE_CHDIR
static rf_int_t fat_cdir(rf_void_t* vol, const rf_char_t* path)
{
    rf_fat_file_ref_t file;
    rf_int_t ret;

    file = file_alloc((rf_volume_comm_ref_t)vol);
    if (file == NULL)
        return RF_EMFILE;

    ret = fat_file_open(file, path, RF_O_RDONLY | RF_O_DIRECTORY);
    if (ret == 0)
        ret = fat_volume_set_current_dir(vol, &file->entry);
    file_free(file);
    return ret;
}
#endif

rf_int_t fat_ioctl_vol(struct rf_vol_ctx* vctx, rf_int_t cmd, rf_void_t* args)
{
    switch (cmd) {
    case RF_VOL_IOCTL_STATFS:
        return fat_volume_stat(vctx->vol, (rf_statfs_ref_t)args);

    case RF_VOL_IOCTL_CHDIR:
        return fat_cdir(vctx->vol, (const rf_char_t*)args);

    }
    return  RF_EPERM;
}

#if RF_OPT_USE_RAWOPEN
rf_int_t fat_raw_open(struct rf_file_ctx* fd, rf_fentry_t* fentry, rf_int_t flags)
{
    rf_int_t ret;
    rf_fat_file_ref_t file;

    file = file_alloc((rf_fat_volume_ref_t)fd->vctx->vol);
    if (file == NULL)
        return RF_EMFILE;

    ret = fat_file_raw_open(file, fentry, flags);
    if (ret != RF_ESUCCESS) {
        file_free(file);
        file = NULL;
    }
    fd->file = file;
    return ret;
}
#endif

rf_int_t fat_open(struct rf_file_ctx* fd, const rf_char_t* file_name, rf_int_t flags)
{
    rf_fat_file_ref_t file;
    rf_int_t ret;

    file = file_alloc((rf_volume_comm_ref_t)fd->vctx->vol);
    if (file == NULL)
        return RF_EMFILE;

    ret = fat_file_open(file, file_name, flags);
    if (ret != RF_ESUCCESS) {
        file_free(file);
        file = NULL;
    }
    fd->file = file;
    return ret;
}

rf_int_t fat_close(struct rf_file_ctx* fd)
{
    rf_int_t ret;

    ret = fat_file_close((rf_file_comm_ref_t)fd->file);
    if (!ret)
        file_free((rf_file_comm_ref_t)fd->file);
    return ret;
}

rf_int_t fat_ioctl(struct rf_file_ctx* fd, rf_int_t cmd, rf_void_t* args)
{
    switch (cmd & 0xff) {
    case RF_FD_IOCTL_GET_FSTAT:
        return fat_file_get_stat(fd->file, (rf_stat_ref_t)args);
#if RF_OPT_USE_GET_FTIME
    case RF_FD_IOCTL_GET_FTIME:
        return fat_file_get_ftime(fd->file, (rf_file_ftime_ref_t)args);
#endif
#if RF_OPT_USE_GET_LTIME
    case RF_FD_IOCTL_GET_LTIME:
        return fat_file_get_ltime(fd->file, (rf_file_ltime_ref_t)args);
#endif
#if RF_OPT_USE_GET_FNAME
    case RF_FD_IOCTL_GET_SHORT_NAME:
        return fat_file_get_name(fd->file, args, (rf_u32_t)(cmd >> 8));
#endif
#if RF_OPT_USE_GET_SET_LONG_NAME
    case RF_FD_IOCTL_GET_NAME:
        return fat_file_get_long_name(fd->file, (rf_wchar_ref_t)args, (rf_u32_t)(cmd >> 8));
#endif
#if RF_OPT_USE_SET_ATTRIB
    case RF_FD_IOCTL_SET_ATTRIB:
        return fat_file_set_attribute(fd->file, *(rf_u32_t*)args);
#endif
#if RF_OPT_USE_SET_FTIME
    case RF_FD_IOCTL_SET_FTIME:
        return fat_file_set_ftime(fd->file, (rf_file_ftime_ref_t)args);
#endif
#if RF_OPT_USE_SET_LTIME
    case RF_FD_IOCTL_SET_LTIME:
        return fat_file_set_ltime(fd->file, (rf_file_ltime_ref_t)args);
#endif
    case RF_FD_IOCTL_FSYNC:
        return fat_file_sync(fd->file);
#if RF_OPT_USE_FTRUNCATE
    case RF_FD_IOCTL_FTRUNCATE:
        return fat_file_truncate(fd->file, (rf_u32_t)(*(rf_fsize_t*)args));
#endif
#if RF_OPT_USE_FDELETE
    case RF_FD_IOCTL_FDELETE:
        return fat_file_delete(fd->file);
#endif
    }
    return RF_EPERM;
}

#if RF_OPT_USE_READDIR
rf_int_t fat_readdir(struct rf_file_ctx* fd, const rf_char_t* pattern, rf_dirent_t* dirp)
{
#if RF_OPT_READDIR_LONG_NAME
    return fat_file_read_dir(fd->file, pattern, dirp, 0);
#else

    return fat_file_read_dir(fd->file, pattern, dirp, RF_O_READDIR_SFN);
#endif
}
#endif

#if RF_OPT_USE_FSEEK
rf_int_t fat_lseek(struct rf_file_ctx* fd, rf_fsize_t offset, rf_int_t origin, rf_fsize_t* now_offset)
{
    rf_int_t ret;

    ret = fat_file_lseek(fd->file, (rf_u32_t)offset, origin);
    if (ret)
        return ret;
    * now_offset = (rf_fsize_t)((rf_fat_file_ref_t)fd->file)->current_offset;
    return 0;
}
#endif

rf_int_t fat_read(struct rf_file_ctx* fd, rf_void_t* buf, rf_size_t len, rf_size_t* plen)
{
    return fat_file_read(fd->file, buf, (rf_u32_t)len, (rf_u32_t*)plen);
}

#if RF_OPT_FS_READONLY == 0
rf_int_t fat_write(struct rf_file_ctx* fd, const rf_void_t* buf, rf_size_t len, rf_size_t* plen)
{
    return fat_file_write(fd->file, (rf_void_t*)buf, (rf_u32_t)len, (rf_u32_t*)plen);
}
#endif

#if RF_OPT_USE_UNLINK
rf_int_t fat_unlink(struct rf_vol_ctx* vctx, const rf_char_t* pathname)
{
    rf_int_t ret, ret1;
    rf_fat_file_ref_t file;

    file = file_alloc((rf_volume_comm_ref_t)vctx->vol);
    if (file == NULL)
        return RF_EMFILE;

    ret = fat_file_open(file, pathname, RF_O_WRONLY);
    if (ret != RF_ESUCCESS) {
        file_free(file);
        return ret;
    }

    ret = fat_file_delete(file);
    RF_ASSERTE("err: fat_file_delete\n", ret == 0);
    ret1 = fat_file_close(file);
    RF_ASSERTE("err: fat_file_delete\n", ret1 == 0);
    file_free(file);
    return ret;
}
#endif

#if RF_OPT_USE_RENAME
rf_int_t fat_rename(struct rf_vol_ctx* vctx, const rf_char_t* oldname, const rf_char_t* newname)
{
    rf_int_t ret;
    rf_fat_file_ref_t oldfile = NULL;
    rf_fat_file_ref_t newfile = NULL;

    do {
        oldfile = file_alloc((rf_volume_comm_ref_t)vctx->vol);
        if (oldfile == NULL) {
            ret = RF_EMFILE;
            break;
        }

        newfile = file_alloc((rf_volume_comm_ref_t)vctx->vol);
        if (newfile == NULL) {
            ret = RF_EMFILE;
            break;
        }

        ret = fat_file_rename(oldfile, newfile, oldname, newname);

    } while (FALSE);

    if (oldfile != NULL) {
        fat_file_close(oldfile);
        file_free(oldfile);
    }
    if (newfile != NULL) {
        fat_file_close(newfile);
        file_free(newfile);
    }
    return ret;
}
#endif

