/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "fat_volume.h"
#include "fat_file.h"

#define N_ROOTDIR           512     /* Number of root directory entries for FAT12/16 */

#define FS_FAT12            1
#define FS_FAT16            2
#define FS_FAT32            3

#define SZ_DIRE             32      /* Size of a directory entry */
/* FAT sub-type boundaries (Differ from specs but correct for real DOS/Windows) */
#define MIN_FAT16           4086U   /* Minimum number of clusters of FAT16 */
#define MIN_FAT32           65526U  /* Minimum number of clusters of FAT32 */

#define BS_jmpBoot          0       /* x86 jump instruction (3) */
#define BS_OEMName          3       /* OEM name (8) */
#define BPB_BytsPerSec      11      /* Sector size [byte] (2) */
#define BPB_SecPerClus      13      /* Cluster size [sector] (1) */
#define BPB_RsvdSecCnt      14      /* Size of reserved area [sector] (2) */
#define BPB_NumFATs         16      /* Number of FAT copies (1) */
#define BPB_RootEntCnt      17      /* Number of root directory entries for FAT12/16 (2) */
#define BPB_TotSec16        19      /* Volume size [sector] (2) */
#define BPB_Media           21      /* Media descriptor (1) */
#define BPB_FATSz16         22      /* FAT size [sector] (2) */
#define BPB_SecPerTrk       24      /* Track size [sector] (2) */
#define BPB_NumHeads        26      /* Number of heads (2) */
#define BPB_HiddSec         28      /* Number of special hidden sectors (4) */
#define BPB_TotSec32        32      /* Volume size [sector] (4) */
#define BS_DrvNum           36      /* Physical drive number (1) */
#define BS_NTres            37      /* Error flag (1) */
#define BS_BootSig          38      /* Extended boot signature (1) */
#define BS_VolID            39      /* Volume serial number (4) */
#define BS_VolLab           43      /* Volume label (8) */
#define BS_FilSysType       54      /* File system type (1) */
#define BPB_FATSz32         36      /* FAT size [sector] (4) */
#define BPB_ExtFlags        40      /* Extended flags (2) */
#define BPB_FSVer           42      /* File system version (2) */
#define BPB_RootClus        44      /* Root directory first cluster (4) */
#define BPB_FSInfo          48      /* Offset of FSINFO sector (2) */
#define BPB_BkBootSec       50      /* Offset of backup boot sector (2) */
#define BS_DrvNum32         64      /* Physical drive number (1) */
#define BS_NTres32          65      /* Error flag (1) */
#define BS_BootSig32        66      /* Extended boot signature (1) */
#define BS_VolID32          67      /* Volume serial number (4) */
#define BS_VolLab32         71      /* Volume label (8) */
#define BS_FilSysType32     82      /* File system type (1) */
#define FSI_LeadSig         0       /* FSI: Leading signature (4) */
#define FSI_StrucSig        484     /* FSI: Structure signature (4) */
#define FSI_Free_Count      488     /* FSI: Number of free clusters (4) */
#define FSI_Nxt_Free        492     /* FSI: Last allocated cluster (4) */
#define MBR_Table           446     /* MBR: Partition table offset (2) */
#define SZ_PTE              16      /* MBR: Size of a partition table entry */
#define BS_55AA             510     /* Signature word (2) */

#if RF_OPT_USE_MKFS

#if RF_OPT_USE_MALLOC
/* fat ڴ ڿٸʽ */
static rf_void_t* max_malloc(rf_int_t* size)
{
    rf_void_t* p;
    for (;;) {
        p = malloc(*size);
        if (p != NULL)
            return p;
        *size >>= 1;
        if (*size == 0)
            return NULL;
    }
}
#endif

rf_int_t fat_mkfs_imp(const rf_char_t* devname, rf_fat_volume_t* vol, const rf_fat_mkfs_param_t* param)
{
    rf_int_t ret;
    /* md = 0xF8 is the standard value for fixed (non-removable) media. For removable media, 0xF0 is frequently used*/
    rf_u8_t fmt, md = 0xf0, sys, *tbl;
    rf_u32_t i, sector_bytes, au;
    rf_u32_t b_fat, s_fat, n_fat;
    rf_u32_t n_clst, vs, n, wsect;
    rf_u32_t n_rsv, b_dir, n_dir, b_data;
    rf_lloff_t ss_count;
    rf_void_t* sector_buffer = NULL;
#if RF_OPT_USE_MALLOC
    rf_int_t tmp_size;
    rf_void_t* tmp_buffer = NULL;
#endif
    rf_handle_t dev_handle;
    rf_mount_ctx_t open_ctx;
    static const rf_u32_t vst[] = { 1024,   512,  256,  128,   64,    32,   16,    8,    4,    2,   0 };
    static const rf_u32_t cst[] = { 32768, 16384, 8192, 4096, 2048, 16384, 8192, 4096, 2048, 1024, 512 };

    /* 򿪴豸 */
    memset(&open_ctx, 0, sizeof(open_ctx));
    open_ctx.volume_offset = param->volume_offset;
    open_ctx.sector_size = param->sector_size;
    ret = rf_disk_open(devname, RF_O_RDWR, &open_ctx, &dev_handle);
    if (ret)
        return ret;

    /* fatֻ 12 */
    s_fat = param->fat_size;
    if (s_fat == 0)
        s_fat = 1;
    if (s_fat > 2)
        s_fat = 2;
    do {
        /* ֽΪ0 ԶȡС С */
        sector_bytes = param->sector_size;
        if (sector_bytes == 0)
            ret = rf_disk_ioctl(dev_handle, RF_CMD_GET_SECTOR_SIZE, &sector_bytes, sizeof(sector_bytes));
        if (ret)
            break;
        if (sector_bytes != 512 && sector_bytes != 1024 && sector_bytes != 2048 && sector_bytes != 4096) {
            ret = RF_ERROR;
            break;
        }

        /* 0Ϊ̬ȡ豸  */
        ss_count = param->volume_size;
        if (ss_count == 0) {
            ret = rf_disk_ioctl(dev_handle, RF_CMD_GET_SECTOR_COUNT, &ss_count, sizeof(ss_count));
            if (ret)
                break;
            /* õС */
            ss_count -= param->volume_offset;
        }

        /* ȡ д */
#if RF_OPT_USE_MALLOC
        sector_buffer = malloc(sector_bytes);
        if (sector_buffer == NULL) {
            ret = RF_ENOMEM;
            break;
        }
#else
        if (RF_OPT_MAX_SECTOR_SIZE < sector_bytes) {
            ret = RF_EBUFSMALL;
            break;
        }
        sector_buffer = vol->sector_buffer;
#endif

        /* ѡȡĴشС */
        au = param->cluster_size;
        if (au & (au - 1))
            au = 0;
        if (!au) {
            vs = (rf_u32_t)(ss_count / (2000 / (sector_bytes / 512)));
            for (i = 0; vs < vst[i]; i++);
            au = cst[i];
        }
        if (au >= sector_bytes)
            au /= sector_bytes; /* Number of sectors per cluster */
        if (!au)
            au = 1;
        if (au > 128)
            au = 128;

        /*  ѡʵļϵͳ */
        n_clst = (rf_u32_t)(ss_count / au);
        /* ݴ ȡfat*/
        fmt = param->fat_type;
        if (fmt == 0) {
            fmt = FS_FAT12;
            if (n_clst >= MIN_FAT16)
                fmt = FS_FAT16;
            if (n_clst >= MIN_FAT32)
                fmt = FS_FAT32;
        }

        /* Determine offset and size of FAT structure */
        if (fmt == FS_FAT32) {
            n_fat = ((n_clst * 4) + 8 + sector_bytes - 1) / sector_bytes;
            n_rsv = 32;
            n_dir = 0;
        } else {
            n_fat = (fmt == FS_FAT12) ? (n_clst * 3 + 1) / 2 + 3 : (n_clst * 2) + 4;
            n_fat = (n_fat + sector_bytes - 1) / sector_bytes;
            n_rsv = 1;
            n_dir = (rf_u32_t)N_ROOTDIR * SZ_DIRE / sector_bytes;
        }
        b_fat = n_rsv;
        b_dir = b_fat + n_fat * s_fat;
        b_data = b_dir + n_dir;
        if (ss_count < b_data + au) {
            ret = RF_ERROR;    /* Too small volume */
            break;
        }

        /* Align data start sector to erase block boundary (for flash memory media) */
        if (rf_disk_ioctl(dev_handle, RF_CMD_GET_BLOCK_SIZE, &n, sizeof(n)) != 0 || !n || n > 32768)
            n = 1;
        n = (b_data + n - 1) & ~(n - 1);    /* Next nearest erase block from current data start */
        n = (n - b_data) / s_fat;
        if (fmt == FS_FAT32) {      /* FAT32: Move FAT offset */
            n_rsv += n;
            b_fat += n;
        } else {                /* FAT12/16: Expand FAT size */
            n_fat += n;
        }

        /* Determine number of clusters and final check of validity of the FAT sub-type */
        n_clst = (rf_u32_t)((ss_count - n_rsv - n_fat * s_fat - n_dir) / au);
        if ((fmt == FS_FAT16 && n_clst < MIN_FAT16) || (fmt == FS_FAT32 && n_clst < MIN_FAT32))
            RF_PRINTF(RF_DBG_LEVEL_WARNING, ("Cluster number is too less\n"));

        /* Determine system ID in the partition table */
        if (fmt == FS_FAT32) {
            sys = 0x0C;     /* FAT32X */
        } else {
            if (fmt == FS_FAT12 && ss_count < 0x10000) {
                sys = 0x01; /* FAT12(<65536) */
            } else {
                sys = (ss_count < 0x10000) ? 0x04 : 0x06;   /* FAT16(<65536) : FAT12/16(>=65536) */
            }
        }

        /* Create BPB in the VBR */
        tbl = (rf_u8_t*)sector_buffer;
        memset(tbl, 0, sector_bytes);
        memcpy(tbl, "\xEB\xFE\x90" "MSDOS5.0", 11);/* Boot jump code, OEM name */
        i = sector_bytes;                               /* Sector size */
        u16_write(tbl, BPB_BytsPerSec, i);
        tbl[BPB_SecPerClus] = (rf_u8_t)au;          /* Sectors per cluster */
        u16_write(tbl, BPB_RsvdSecCnt, n_rsv);  /* Reserved sectors */
        tbl[BPB_NumFATs] = s_fat;               /* Number of FATs */
        i = (fmt == FS_FAT32) ? 0 : N_ROOTDIR;  /* Number of root directory entries */
        u16_write(tbl, BPB_RootEntCnt, i);
        if (ss_count < 0x10000) {                   /* Number of total sectors */
            u16_write(tbl, BPB_TotSec16, ss_count);
        } else {
            u32_write(tbl, BPB_TotSec32, ss_count);
        }
        tbl[BPB_Media] = md;                    /* Media descriptor */
        u16_write(tbl, BPB_SecPerTrk, 63);      /* Number of sectors per track */
        u16_write(tbl, BPB_NumHeads, 255);      /* Number of heads */
        u32_write(tbl, BPB_HiddSec, param->volume_offset);      /* Hidden sectors */
        u8_write(tbl, MBR_Table, 0x07);
        n = fat_time_current();                     /* Use current time as VSN */
        if (fmt == FS_FAT32) {
            u32_write(tbl, BS_VolID32, n);      /* VSN */
            u32_write(tbl, BPB_FATSz32, n_fat); /* Number of sectors per FAT */
            u32_write(tbl, BPB_RootClus, 2);    /* Root directory start cluster (2) */
            u16_write(tbl, BPB_FSInfo, 1);      /* FSINFO record offset (VBR + 1) */
            u16_write(tbl, BPB_BkBootSec, 6);   /* Backup boot record offset (VBR + 6) */
            tbl[BS_DrvNum32] = 0x80;            /* Drive number */
            tbl[BS_BootSig32] = 0x29;           /* Extended boot signature */
            memcpy(tbl + BS_VolLab32, "NO NAME    " "FAT32   ", 19);    /* Volume label, FAT signature */
        } else {
            u32_write(tbl, BS_VolID, n);        /* VSN */
            u16_write(tbl, BPB_FATSz16, n_fat); /* Number of sectors per FAT */
            tbl[BS_DrvNum] = 0x80;              /* Drive number */
            tbl[BS_BootSig] = 0x29;             /* Extended boot signature */
            memcpy(tbl + BS_VolLab, "NO NAME    " "FAT     ", 19);  /* Volume label, FAT signature */
        }
        u16_write(tbl, BS_55AA, 0xAA55);            /* Signature (Offset is fixed here regardless of sector size) */
        ret = rf_disk_write(dev_handle, 0, 1, tbl); /* Write it to the VBR sector */
        if (ret)
            break;
        if (fmt == FS_FAT32)                    /* Write it to the backup VBR if needed (VBR + 6) */
            rf_disk_write(dev_handle, 6, 1, tbl);

        /* Initialize FAT area */
        wsect = b_fat;
        for (i = 0; i < s_fat; i++) {       /* Initialize each FAT copy */
            memset(tbl, 0, sector_bytes);
            n = md;                             /* Media descriptor byte */
            if (fmt != FS_FAT32) {
                n |= (fmt == FS_FAT12) ? 0x00FFFF00 : 0xFFFFFF00;
                u32_write(tbl, 0, n);           /* Reserve cluster #0-1 (FAT12/16) */
            } else {
                n |= 0xFFFFFF00;
                u32_write(tbl, 0, n);           /* Reserve cluster #0-1 (FAT32) */
                u32_write(tbl, 4, 0xFFFFFFFF);
                u32_write(tbl, 8, 0x0FFFFFFF);  /* Reserve cluster #2 for root directory */
            }
            ret = rf_disk_write(dev_handle, wsect++, 1, tbl);
            if (ret)
                goto _exit;

            if (n_fat == 1)
                continue;
#if RF_OPT_USE_MALLOC
            tmp_size = (n_fat - 1) * sector_bytes;
            if (tmp_buffer == NULL) {
                tmp_buffer = max_malloc(&tmp_size);
                if (tmp_buffer == NULL)
                    return RF_ENOMEM;
                memset(tmp_buffer, 0, tmp_size);
            }
            tmp_size /= sector_bytes;
            for (n = 0; n < (n_fat - 1) / tmp_size; n++) {
                ret = rf_disk_write(dev_handle, wsect, tmp_size, tmp_buffer);
                if (ret)
                    goto _exit;
                wsect += tmp_size;
            }
#else
            memset(tbl, 0, sector_bytes);
            for (n = 1; n < n_fat; n++) {
                ret = rf_disk_write(dev_handle, wsect++, 1, tbl);
                if (ret)
                    goto _exit;
            }
#endif
        }

        /* Initialize root directory */
        i = (fmt == FS_FAT32) ? au : (rf_u32_t)n_dir;
#if RF_OPT_USE_MALLOC
        if (tmp_buffer == NULL) {
            tmp_size = i * sector_bytes;
            tmp_buffer = max_malloc(&tmp_size);
            if (tmp_buffer == NULL)
                return RF_ENOMEM;
            memset(tmp_buffer, 0, tmp_size);
            tmp_size /= sector_bytes;
        }
        for (n = 0; n < i / tmp_size; n++) {
            ret = rf_disk_write(dev_handle, wsect, tmp_size, tmp_buffer);
            if (ret)
                goto _exit;
            wsect += tmp_size;
        }
        /* дβ󼸸 */
        i -= (i / tmp_size) * tmp_size;
        if (i) {
            ret = rf_disk_write(dev_handle, wsect, i, tmp_buffer);
            if (ret)
                goto _exit;
        }
#else
        do {
            ret = rf_disk_write(dev_handle, wsect++, 1, tbl);
            if (ret)
                goto _exit;
        } while (--i);
#endif

        if (fmt == FS_FAT32) {
            u32_write(tbl, FSI_LeadSig, 0x41615252);
            u32_write(tbl, FSI_StrucSig, 0x61417272);
            u32_write(tbl, FSI_Free_Count, n_clst - 1); /* Number of free clusters */
            u32_write(tbl, FSI_Nxt_Free, 2);            /* Last allocated cluster# */
            u16_write(tbl, BS_55AA, 0xAA55);
            ret = rf_disk_write(dev_handle, param->volume_offset + 1, 1, tbl); /* Write original (VBR + 1) */
            ret = rf_disk_write(dev_handle, param->volume_offset + 7, 1, tbl); /* Write backup (VBR + 7) */
        }

        ret = rf_disk_ioctl(dev_handle, RF_CMD_CTRL_SYNC, NULL, 0);
    } while (0);
_exit:
#if RF_OPT_USE_MALLOC
    if (sector_buffer != NULL)
        rf_free(sector_buffer);
    if (tmp_buffer != NULL)
        rf_free(tmp_buffer);
#endif
    rf_disk_close(dev_handle);
    return ret;
}
#endif
