/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "ntfs_volume.h"
#include "ntfs_file.h"
#include "ntfs.h"

static rf_int_t is_boot_sector_ntfs(rf_u8_t* sector_buffer)
{
    rf_int_t i;

    /* FIXME: We don't use checksumming yet as NT4(SP6a) doesn't either...
     * But we might as well have the code ready to do it. (AIA) */
#if 0
    /* Calculate the checksum. */
    if (b < b + 0x50) {
        rf_u32_t* u;
        rf_u32_t* bi = (rf_u32_t*)(b + 0x50);

        for (u = bi, i = 0; u < bi; ++u)
            i += NTFS_GETU32(*u);
    }
#endif
    /* Check magic is "NTFS    " */
    if (sector_buffer[3] != 0x4e)
        goto not_ntfs;
    if (sector_buffer[4] != 0x54)
        goto not_ntfs;
    if (sector_buffer[5] != 0x46)
        goto not_ntfs;
    if (sector_buffer[6] != 0x53)
        goto not_ntfs;
    for (i = 7; i < 0xb; ++i) {
        if (sector_buffer[i] != 0x20)
            goto not_ntfs;
    }

    /* Check bytes per sector value is between 512 and 4096. */
    if (sector_buffer[0xb] != 0)
        goto not_ntfs;
    if (sector_buffer[0xc] > 0x10)
        goto not_ntfs;
    /* Check sectors per cluster value is valid. */
    switch (sector_buffer[0x0d]) {
    case 1:
    case 2:
    case 4:
    case 8:
    case 16:
    case 32:
    case 64:
    case 128:
        break;
    default:
        goto not_ntfs;
    }

    /* Check reserved sectors value and four other fields are zero. */
    for (i = 0xe; i < 0x15; ++i) {
        if (sector_buffer[i] != 0)
            goto not_ntfs;
    }
    if (sector_buffer[0x16] != 0)
        goto not_ntfs;
    if (sector_buffer[0x17] != 0)
        goto not_ntfs;
    for (i = 0x20; i < 0x24; ++i) {
        if (sector_buffer[i] != 0)
            goto not_ntfs;
    }

    /* Check clusters per file record segment value is valid. */
    if (sector_buffer[0x40] < 0xe1 || sector_buffer[0x40] > 0xf7) {
        switch (sector_buffer[0x40]) {
        case 1:
        case 2:
        case 4:
        case 8:
        case 16:
        case 32:
        case 64:
            break;
        default:
            goto not_ntfs;
        }
    }

    /* Check clusters per index block value is valid. */
    if (sector_buffer[0x44] < 0xe1 || sector_buffer[0x44] > 0xf7) {
        switch (sector_buffer[0x44]) {
        case 1:
        case 2:
        case 4:
        case 8:
        case 16:
        case 32:
        case 64:
            break;
        default:
            goto not_ntfs;
        }
    }
    return RF_ESUCCESS;

not_ntfs:
    return RF_ENOFS;
}

static rf_int_t ntfs_volume_get_mft_block(rf_ntfs_volume_ref_t vol)
{
    rf_int_t ret, i;
    rf_u16_t idx_offset = 0;
    rf_u8_t tmp;

    /* ȶȡMFTļ,Ҫע20150410Ҫȡһ */
    ret = ntfs_read_attr(vol, (vol->mft_lcn << vol->sectors_per_cluster_bits),
                         TRUE, AT_DATA, NULL, &idx_offset);
    if (ret)
        return ret;

    idx_offset += 0x40;  /* ˴0x40 ΪԵΪ0Դ0x40ʼȡRun listϢ */
    for (i = 0; i < MAX_MFT_BLOCK; i++) {
        tmp = u8_read(vol->sector_buffer, idx_offset++);   /*һmftС ȶDataRunNLڡݻָȽءP310 */
        vol->mft_size[i] = get_s32(vol->sector_buffer, (rf_u16_t)idx_offset, tmp & 0x0f);   /*  2019-5-5 ԭΪvol->mft_offset */

        idx_offset += tmp & 0x0f;
        vol->mft_offet[i] = get_s32(vol->sector_buffer, (rf_u16_t)idx_offset, tmp >> 4); /*  2019-5-5 ԭΪvol->mft_size */
        if(i>0)
            vol->mft_offet[i]+= vol->mft_offet[i-1]; /* 2019-5-6 Ӹ */
        idx_offset += tmp >> 4;
        if (u8_read(vol->sector_buffer, idx_offset) == 0)
            break;
    }

    if (u8_read(vol->sector_buffer, idx_offset) != 0) {
        return RF_ERROR;
    }
    return RF_ESUCCESS;
}

rf_int_t ntfs_volume_open(rf_ntfs_volume_ref_t vol, rf_mount_param_t* param)
{
    rf_s32_t sector_bytes, mft_zone_size, tc;
    rf_u64_t ll;
    rf_int_t ret;

    ret = rf_disk_open(param->device_name, param->flags, param->ctx, &vol->dev_handle);
    if (ret)
        return ret;
    vol->access_flag = 0;
    if ((param->flags & 0x3) == 2)
        vol->access_flag = 1;

    ret = rf_disk_ioctl(vol->dev_handle, RF_CMD_GET_SECTOR_SIZE, &sector_bytes, sizeof(sector_bytes));
    if (ret)
        return ret;
    RF_PERROR("sector bytes error", INVOID_SECTOR_SIZE(sector_bytes), return RF_ENOFS);

#if RF_OPT_USE_MALLOC
    /* ntfs С仺 1024ֽ */
    vol->sector_buffer = malloc(max(sector_bytes, 1024));
    if (vol->sector_buffer == NULL)
        return RF_ENOMEM;
#else
    if (RF_OPT_MAX_SECTOR_SIZE < sector_bytes)
        return RF_EBUFSMALL;
#endif

    ret = rf_disk_read(vol->dev_handle, 0, 1, vol->sector_buffer);
    if (ret)
        return ret;
    if (u16_read(vol->sector_buffer, 510) != 0xAA55)
        return RF_ENOFS;
    ret = is_boot_sector_ntfs(vol->sector_buffer);
    if (ret != RF_ESUCCESS)
        return ret;

    vol->sector_size = (rf_u16_t)sector_bytes;
    vol->sector_size_bits = rf_ffs(sector_bytes);
    vol->mft_sizeof_sector = 2;
    if (vol->sector_size >= 1024)
        vol->mft_sizeof_sector = 1;

    vol->sectors_per_cluster = u8_read(vol->sector_buffer, 0xD);
    vol->sectors_per_cluster_bits = rf_ffs(vol->sectors_per_cluster);
    vol->mft_clusters_per_record = u8_read(vol->sector_buffer, 0x40);
    vol->index_clusters_per_record = u8_read(vol->sector_buffer, 0x44);
    vol->cluster_size = vol->sector_size << vol->sectors_per_cluster_bits;
    vol->cluster_size_bits = rf_ffs(vol->cluster_size);
    if (vol->mft_clusters_per_record > 0) {
        vol->mft_record_size = vol->cluster_size << (rf_ffs(vol->mft_clusters_per_record));
    } else {
        /*
         * When mft_record_size < cluster_size, mft_clusters_per_record
         * = -log2(mft_record_size) bytes. mft_record_size normaly is
         * 1024 bytes, which is encoded as 0xF6 (-10 in decimal).
         */
        vol->mft_record_size = 1 << -vol->mft_clusters_per_record;
    }

    vol->mft_record_size_bits = rf_ffs(vol->mft_record_size);
    if (vol->index_clusters_per_record > 0) {
        vol->index_record_size = vol->cluster_size << rf_ffs(vol->index_clusters_per_record);
    } else {
        /*
        * When index_record_size < cluster_size,
        * index_clusters_per_record = -log2(index_record_size) bytes.
        * index_record_size normaly equals 4096 bytes, which is
        * encoded as 0xF4 (-12 in decimal).
        */
        vol->index_record_size = 1 << -vol->index_clusters_per_record;
    }

    vol->index_record_size_bits = rf_ffs(vol->index_record_size);
    /*
     * Get the size of the volume in clusters (ofs 0x28 is nr_sectors) and
     * check for 64-bit-ness. Windows currently only uses 32 bits to save
     * the clusters so we do the same as it is much faster on 32-bit CPUs.
     */
    ll = u64_read(vol->sector_buffer, 0x28) >> vol->sectors_per_cluster_bits;
    if (ll >= (rf_s64_t)1 << 31)
        return RF_ENOFS;

    vol->total_clusters = (rf_s32_t)ll;
    vol->mft_lcn = (rf_s32_t)u64_read(vol->sector_buffer, 0x30);
    vol->mft_mirr_lcn = (rf_s32_t)u64_read(vol->sector_buffer, 0x38);
    /* Determine MFT zone size. */
    mft_zone_size = vol->total_clusters;
    switch (vol->mft_zone_multiplier) {
    /* % of volume size in clusters */
    case 4:
        mft_zone_size >>= 1;            /* 50%   */
        break;
    case 3:
        mft_zone_size = mft_zone_size * 3 >> 3; /* 37.5% */
        break;
    case 2:
        mft_zone_size >>= 2;            /* 25%   */
        break;
    /* case 1: */
    default:
        mft_zone_size >>= 3;            /* 12.5% */
        break;
    }
    /* Setup mft zone. */
    vol->mft_zone_start = vol->mft_zone_pos = vol->mft_lcn;
    /*
     * Calculate the mft_lcn for an unmodified NTFS volume (see mkntfs
     * source) and if the actual mft_lcn is in the expected place or even
     * further to the front of the volume, extend the mft_zone to cover the
     * beginning of the volume as well. This is in order to protect the
     * area reserved for the mft bitmap as well within the mft_zone itself.
     * On non-standard volumes we don't protect it as well as the overhead
     * would be higher than the speed increase we would get by doing it.
     */
    tc = (8192 + 2 * vol->cluster_size - 1) / vol->cluster_size;
    if (tc * vol->cluster_size < 16 * 1024)
        tc = (16 * 1024 + vol->cluster_size - 1) / vol->cluster_size;
    if (vol->mft_zone_start <= tc)
        vol->mft_zone_start = (rf_s32_t)0;
    /*
     * Need to cap the mft zone on non-standard volumes so that it does
     * not point outside the boundaries of the volume, we do this by
     * halving the zone size until we are inside the vol->
     */
    vol->mft_zone_end = vol->mft_lcn + mft_zone_size;
    while (vol->mft_zone_end >= vol->total_clusters) {
        mft_zone_size >>= 1;
        vol->mft_zone_end = vol->mft_lcn + mft_zone_size;
    }

    /*
     * Set the current position within each data zone to the start of the
     * respective zone.
     */
    vol->data1_zone_pos = vol->mft_zone_end;
    vol->data2_zone_pos = (rf_s32_t)0;
    ret = rf_disk_read(vol->dev_handle, (vol->mft_lcn << vol->sectors_per_cluster_bits),
                       vol->mft_sizeof_sector, vol->sector_buffer);
    if (ret)
        return ret;

    ret = ntfs_fixup_record(vol, "FILE");
    if (ret)
        return RF_ERROR;
    if (vol->mft_sizeof_sector == 2)
        vol->root_dir_entry.sector = (vol->mft_lcn << vol->sectors_per_cluster_bits) + (5 << 1);
    else
        vol->root_dir_entry.sector = (vol->mft_lcn << vol->sectors_per_cluster_bits) + 5;
    /* ĬϵǰĿ¼ ΪĿ¼ */
    vol->current_dir_entry.sector = vol->root_dir_entry.sector;

    ret = ntfs_volume_get_mft_block(vol);
    if (ret == 0)
        vol->fs_type = RF_TYPE_NTFS;
    return ret;
}

rf_int_t ntfs_volume_close(rf_ntfs_volume_ref_t vol)
{
    rf_int_t ret;

    ret = rf_disk_close(vol->dev_handle);
    if (ret == RF_ESUCCESS) {
#if RF_OPT_USE_MALLOC == 1
        if (vol->sector_buffer != NULL)
            rf_free(vol->sector_buffer);
#endif
    }
    return ret;
}

rf_int_t ntfs_volume_stat(rf_ntfs_volume_ref_t vol, rf_statfs_ref_t st)
{
    st->f_type = vol->fs_type;
    st->f_ssize = vol->sector_size;
    st->f_bsize = vol->sectors_per_cluster << vol->sector_size_bits;
    st->f_bfree = 0;
    st->f_blocks = vol->total_clusters;
    return 0;
}

rf_int_t ntfs_volume_get_label(rf_ntfs_volume_ref_t vol, const rf_char_t* buf, rf_int_t len)
{
    return 0;
}

rf_int_t ntfs_volume_set_label(rf_ntfs_volume_ref_t vol, const rf_char_t* buf)
{
    return 0;
}

rf_int_t ntfs_volume_set_current_dir(rf_ntfs_volume_ref_t vol, rf_fentry_t* entry)
{
    vol->current_dir_entry.sector = entry->sector;
    vol->current_dir_entry.offset = entry->offset;
    return 0;
}
