/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "rf_dos.h"
#include <string.h>

/* 512 << 7 = 65536 byte sectors.  */
#define MAX_SECTOR_LOG                  7

#define GPT_PARTITIONS_OFFSET           0x48
#define GPT_MAXPART_OFFSET              0x50
#define GPT_PARTITIONS_SIZE_OFFSET      0x54

#define GPT_START_LBA_OFFSET            0x20
#define GPT_END_LBA_OFFSET              0x28

#if RF_OPT_USE_GPT
static rf_u8_t gpt_magic[8] = {
    0x45, 0x46, 0x49, 0x20, 0x50, 0x41, 0x52, 0x54
};

rf_int_t is_type_empty(rf_u8_t* buffer)
{
    rf_int_t i;

    for (i = 0; i < 16; i++)
        if (*(buffer + i) != '\0')
            return 0;
    return 1;
}

rf_int_t gpt_get_parts(rf_handle_t dev, rf_u8_t* buffer, rf_part_param_t* param)
{
    rf_int_t ret;
    rf_int_t i, sector_log, last_offset = 0, max_part, partentry_size;
    rf_u8_t sys_ind;
    rf_u64_t entries;

    ret = rf_disk_read(dev, 0, 1, buffer);
    if (ret)
        return ret;

    if ((buffer[DOS_PART_MAGIC_OFFSET + 0] != 0x55) ||
            (buffer[DOS_PART_MAGIC_OFFSET + 1] != 0xaa)) {
        return RF_ENOFS;
    }

    for (i = 0; i < 4; i++) {
        sys_ind = buffer[DOS_PART_TBL_OFFSET + (i * 16 + DOS_PART_SYS_IND_OFFSET)];
        if (sys_ind == DOS_PARTITION_TYPE_GPT_DISK)
            break;
    }
    if (i == 4)
        return RF_ENOFS;

    for (sector_log = 0; sector_log < MAX_SECTOR_LOG; sector_log++) {
        ret = rf_disk_read(dev, (rf_lloff_t)1 << sector_log, 1, buffer);
        if (ret)
            return ret;

        if (memcmp(buffer, gpt_magic, sizeof(gpt_magic)) == 0)
            break;
    }
    if (sector_log == MAX_SECTOR_LOG)
        return RF_ENOFS;
    max_part = u32_read(buffer, GPT_MAXPART_OFFSET);
    partentry_size = u32_read(buffer, GPT_PARTITIONS_SIZE_OFFSET);

    entries = u64_read(buffer, GPT_PARTITIONS_OFFSET) << sector_log;
    ret = rf_disk_read(dev, entries, 1, buffer);
    if (ret)
        return ret;

    for (i = 0; i < max_part; i++) {
        if (!is_type_empty(buffer + last_offset)) {
            if (param->count < RF_MAX_DISK_PART) {
                param->part[param->count].start = u64_read(buffer, last_offset + GPT_START_LBA_OFFSET) << sector_log;
                param->part[param->count].size = (u64_read(buffer, last_offset + GPT_END_LBA_OFFSET)
                                                  - u64_read(buffer, last_offset + GPT_START_LBA_OFFSET) + 1) << sector_log;
            }
            param->count++;
        }

        last_offset += partentry_size;
        if (last_offset == param->sector_size) {
            last_offset = 0;
            ret = rf_disk_read(dev, ++entries, 1, buffer);
            if (ret)
                return ret;
        }
    }
    return ret;
}
#endif
