/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "../common/rf_commonx.h"
#include "../src/rf_fs.h"

#if RF_OPT_USE_FAT
#include "../fat/fat_api.h"
#endif
#if RF_OPT_USE_EXT2
#include "../ext2/ext2_api.h"
#endif
#if RF_OPT_USE_NTFS
#include "../ntfs/ntfs_api.h"
#endif

#if RF_OPT_USE_RF_PART
rf_int_t dos_get_parts(rf_handle_t dev, rf_u8_t* buffer, rf_part_param_t* param);
rf_int_t gpt_get_parts(rf_handle_t dev, rf_u8_t* buffer, rf_part_param_t* param);

rf_void_t* rf_alloc_sector_buf(rf_int_t size, rf_int_t* ret)
{
#if RF_OPT_USE_MALLOC
    *ret = RF_ENOMEM;
    return rf_malloc(size);
#else
    rf_vol_comm_ref_t vol;
    if (size > RF_OPT_MAX_SECTOR_SIZE) {
        *ret = RF_EBUFSMALL;
        return NULL;
    }

#if RF_OPT_USE_FAT
    vol = fat_vol_alloc();
    if (vol)
        return vol->sector_buffer;
#endif

#if RF_OPT_USE_NTFS
    vol = ntfs_vol_alloc();
    if (vol)
        return vol->sector_buffer;
#endif

#if RF_OPT_USE_EXT2
    vol = ext2_vol_alloc();
    if (vol)
        return vol->sector_buffer;
#endif

    RF_ASSERTE("error\n", FALSE);
    return NULL;
#endif
}

rf_void_t rf_free_sector_buf(rf_void_t* buf)
{
#if RF_OPT_USE_MALLOC
    rf_free(buf);
#else
    rf_vol_comm_ref_t vol;

    vol = RF_CONTAINING_RECORD(buf, rf_vol_comm_t, sector_buffer);
    switch (vol->fs_type) {
#if RF_OPT_USE_FAT
    case RF_TYPE_FAT32:
    case RF_TYPE_FAT16:
    case RF_TYPE_FAT12:
        fat_vol_free(vol);
        break;
#endif

#if RF_OPT_USE_NTFS
    case RF_TYPE_NTFS:
        ntfs_vol_free(vol);
        break;
#endif

#if RF_OPT_USE_EXT2
    case RF_TYPE_EXT2:
    case RF_TYPE_EXT3:
        ext2_vol_free(vol);
        break;
#endif

    default:
        RF_PERROR("error: invalid fs_type\n", 1, return);
    }
#endif
}

rf_int_t rf_imp_part(rf_char_t* name, rf_part_param_t* param)
{
    rf_int_t ret;
    rf_int_t sector_bytes;
    rf_handle_t dev;
    rf_mount_ctx_t open_ctx;
    rf_u8_t* sector_buffer = NULL;

    memset(&open_ctx, 0, sizeof(open_ctx));
    open_ctx.volume_offset = 0;
    open_ctx.sector_size = param->sector_size;

    ret = rf_disk_open(name, RF_O_RDONLY, &open_ctx, &dev);
    if (ret)
        goto _EXIT;

    do {
        sector_bytes = param->sector_size;
        if (sector_bytes == 0) {
            ret = rf_disk_ioctl(dev, RF_CMD_GET_SECTOR_SIZE, &sector_bytes, sizeof(sector_bytes));
            if (ret)
                break;
        }

        sector_buffer = rf_alloc_sector_buf(sector_bytes, &ret);
        if (sector_buffer == NULL)
            break;

        ret = RF_ENOFS;
#if RF_OPT_USE_GPT
        ret = gpt_get_parts(dev, sector_buffer, param);
        if (ret == 0)
            break;
#endif

#if RF_OPT_USE_MBR
        ret = dos_get_parts(dev, sector_buffer, param);
        if (ret == 0)
            break;
#endif
    } while (FALSE);
    if (sector_buffer != NULL)
        rf_free_sector_buf(sector_buffer);
    rf_disk_close(dev);

_EXIT:
    if (ret != 0) {
        rf_set_errno(ret);
        return -1;
    }
    return 0;
}

#endif
