/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "rf_fs.h"
#include "rf_fd.h"

#if RF_OPT_USE_MALLOC
#define IOINFO_L2E          5
#define IOINFO_ARRAY_ELTS   (1 << IOINFO_L2E)
#define IOINFO_ARRAYS       64
#define _NHANDLE_           (IOINFO_ARRAYS * IOINFO_ARRAY_ELTS)
#define _pioinfo(i)         ( __pfile_ctx[(i) >> IOINFO_L2E] + ((i) & (IOINFO_ARRAY_ELTS - 1)) )
static rf_int_t             _nhandle = 0;
static rf_file_ctx_t*       __pfile_ctx[IOINFO_ARRAYS] = { NULL };
#else
static rf_file_ctx_t        g_file_ctx_table[RF_OPT_MAX_FILE];
#endif

#if RF_OPT_USE_MALLOC
rf_int_t fd_init(rf_void_t)
{
    _nhandle = 0;
    return 0;
}

rf_int_t fd_new(rf_void_t)
{
    rf_int_t i, fh = -1;
    rf_file_ctx_t* pio;

    for (i = 0; i < IOINFO_ARRAYS; i++) {
        if (__pfile_ctx[i] != NULL) {
            for (pio = __pfile_ctx[i]; pio < __pfile_ctx[i] + IOINFO_ARRAY_ELTS; pio++) {
                if (!pio->open_flag) {
                    pio->open_flag = TRUE;
                    fh = i * IOINFO_ARRAY_ELTS + (int)(pio - __pfile_ctx[i]);
                    return fh;
                }
            }
        } else {
            pio = rf_malloc(IOINFO_ARRAY_ELTS * sizeof(rf_file_ctx_t));
            RF_PERROR("error: malloc failed\n", (pio == NULL), return -1);
            memset(pio, 0, IOINFO_ARRAY_ELTS * sizeof(rf_file_ctx_t));
            __pfile_ctx[i] = pio;
            _nhandle += IOINFO_ARRAY_ELTS;
            fh = i * IOINFO_ARRAY_ELTS;
            _pioinfo(i)->open_flag = TRUE;
            break;
        }
    }
    return (fh);
}

rf_file_ctx_t* fd_get_check(rf_int_t fd)
{
    RF_PERROR("error: invalid fd\n", (fd < 0 || fd >= _nhandle), return NULL);
    RF_PERROR("error: fd is not open\n", !_pioinfo(fd)->open_flag, return NULL);
    return _pioinfo(fd);
}

#if RF_OPT_USE_RFS_EXIT
rf_file_ctx_t* fd_get(rf_int_t fd)
{
    RF_PERROR("error: invalid fd\n", (fd < 0 || fd >= _nhandle), return NULL);
    return _pioinfo(fd);
}

rf_int_t fd_get_open_fd(rf_void_t)
{
    rf_int_t i, count = 0;
    rf_file_ctx_t* pio;

    for (i = 0; i < IOINFO_ARRAYS; i++) {
        if (__pfile_ctx[i] != NULL) {
            for (pio = __pfile_ctx[i]; pio < __pfile_ctx[i] + IOINFO_ARRAY_ELTS; pio++)
                if (pio->open_flag)
                    count++;
        }
    }
    return count;
}

rf_int_t fd_get_max_fd(rf_void_t)
{
    return _nhandle;
}

rf_int_t fd_exit(rf_void_t)
{
    rf_int_t i;

    for (i = 0; i < IOINFO_ARRAYS; i++) {
        if (__pfile_ctx[i] != NULL) {
            rf_free(__pfile_ctx[i]);
            __pfile_ctx[i] = NULL;
        }
    }
    return 0;
}
#endif
#else

rf_int_t fd_init(rf_void_t)
{
    memset(&g_file_ctx_table, 0, sizeof(g_file_ctx_table));
    return 0;
}

rf_int_t fd_new(rf_void_t)
{
    rf_file_ctx_t* d;
    rf_int_t idx;

    for (idx = 0; idx < RF_OPT_MAX_FILE && g_file_ctx_table[idx].open_flag; idx++);
    if (idx == RF_OPT_MAX_FILE)
        return -1;
    d = &(g_file_ctx_table[idx]);
    d->open_flag = TRUE;
    return idx;
}

rf_file_ctx_t* fd_get_check(rf_int_t fd)
{
    RF_PERROR("error: invalid fd\n", (fd < 0 || fd >= RF_OPT_MAX_FILE), return NULL);
    RF_PERROR("error: fd is not open\n", !g_file_ctx_table[fd].open_flag, return NULL);
    return &g_file_ctx_table[fd];
}

#if RF_OPT_USE_RFS_EXIT
rf_file_ctx_t* fd_get(rf_int_t fd)
{
    RF_PERROR("error: invalid fd\n", (fd < 0 || fd >= RF_OPT_MAX_FILE), return NULL);
    return &g_file_ctx_table[fd];
}

rf_int_t fd_get_max_fd(rf_void_t)
{
    return RF_OPT_MAX_FILE;
}

rf_int_t fd_get_open_fd(rf_void_t)
{
    rf_int_t i, c;

    for (i = 0, c = 0; i < RF_OPT_MAX_FILE; i++)
        if (g_file_ctx_table[i].open_flag)
            c++;
    return c;
}

rf_int_t fd_exit(rf_void_t)
{
    return 0;
}
#endif
#endif

rf_int_t fd_free(rf_file_ctx_t* ctx)
{
    RF_PERROR("error: ctx is not open\n", !ctx->open_flag, return -1);
    ctx->open_flag = FALSE;
    return 0;
}
