/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "rf_fs.h"
#if RF_OPT_USE_FAT
#include "../fat/fat_api.h"
#endif
#if RF_OPT_USE_NTFS
#include "../ntfs/ntfs_api.h"
#endif
#if RF_OPT_USE_EXT2
#include "../ext2/ext2_api.h"
#endif

#if RF_OPT_USE_RAWOPEN
rf_int_t rf_file_raw_open(rf_file_ctx_t* fctx, const rf_char_t* path, rf_fentry_t* fentry, rf_int_t flags)
{
    rf_vol_ctx_t* vctx;

    if (!fentry)
        return RF_EINVAL;
    vctx = find_fs_ctx(path);
    FS_CHECK(vctx);
    fctx->vctx = vctx;
    switch (((rf_vol_comm_t*)vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        return fat_raw_open(fctx, fentry, flags);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        return ntfs_raw_open(fctx, fentry, flags);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        return ext2_raw_open(fctx, fentry, flags);
        break;
#endif

    default:
        return RF_ERROR;
    }
}
#endif

rf_int_t rf_file_open(rf_file_ctx_t* fctx, const rf_char_t* path, rf_int_t flags)
{
    rf_vol_ctx_t* vctx;

    vctx = find_fs_ctx(path);
    FS_CHECK(vctx);

    fctx->vctx = vctx;
    switch (((rf_vol_comm_t*)vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        return fat_open(fctx, path_get_root_dir(path), flags);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        return ntfs_open(fctx, path_get_root_dir(path), flags);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        return ext2_open(fctx, path_get_root_dir(path), flags);
        break;
#endif

    default:
        return RF_ERROR;
    }
}

rf_int_t rf_file_close(rf_file_ctx_t* fctx)
{
    switch (((rf_vol_comm_t*)fctx->vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        return fat_close(fctx);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        return ntfs_close(fctx);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        return ext2_close(fctx);
        break;
#endif

    default:
        return RF_ERROR;
    }
}

rf_int_t rf_file_ioctl(rf_file_ctx_t* fctx, rf_int_t cmd, rf_void_t* args)
{
    switch (((rf_vol_comm_t*)fctx->vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        return fat_ioctl(fctx, cmd, args);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        return ntfs_ioctl(fctx, cmd, args);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        return ext2_ioctl(fctx, cmd, args);
        break;
#endif

    default:
        return RF_ERROR;
    }
}

#if RF_OPT_USE_FSEEK
rf_int_t rf_file_lseek(rf_file_ctx_t* fctx, rf_off_t offset, rf_int_t origin, rf_off_t* poffset)
{
    switch (((rf_vol_comm_t*)fctx->vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        return fat_lseek(fctx, (rf_fsize_t)offset, origin, (rf_fsize_t*)poffset);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        return ntfs_lseek(fctx, (rf_fsize_t)offset, origin, (rf_fsize_t*)poffset);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        return ext2_lseek(fctx, (rf_fsize_t)offset, origin, (rf_fsize_t*)poffset);
        break;
#endif

    default:
        return RF_ERROR;
    }
}
#endif

rf_int_t rf_file_read(rf_file_ctx_t* fctx, rf_void_t* buf, rf_size_t len, rf_size_t* plen)
{
    rf_int_t ret;
    rf_size_t rlen = 0;

    switch (((rf_vol_comm_t*)fctx->vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        ret = fat_read(fctx, buf, len, &rlen);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        ret = ntfs_read(fctx, buf, len, &rlen);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        ret = ext2_read(fctx, buf, len, &rlen);
        break;
#endif

    default:
        return RF_ERROR;
    }
    if (ret == RF_ESUCCESS)
        *plen = rlen;
    return ret;
}

#if RF_OPT_USE_READDIR
rf_int_t rf_file_readdir(rf_file_ctx_t* fctx, const rf_char_t* pattern, rf_dirent_t* dirp)
{
    if (!pattern || !dirp)
        return RF_EINVAL;
    switch (((rf_vol_comm_t*)fctx->vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        return fat_readdir(fctx, pattern, dirp);
        break;
#endif

#if RF_OPT_USE_NTFS
        FS_NTFS_TYPE
        return ntfs_readdir(fctx, pattern, dirp);
        break;
#endif

#if RF_OPT_USE_EXT2
        FS_EXT_TYPE
        return ext2_readdir(fctx, pattern, dirp);
        break;
#endif

    default:
        return RF_ERROR;
    }
}
#endif

rf_int_t rf_file_write(rf_file_ctx_t* fctx, const rf_void_t* buf, rf_size_t len, rf_size_t* plen)
{
    rf_size_t wlen = 0;
    rf_int_t ret;

    switch (((rf_vol_comm_t*)fctx->vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        ret = fat_write(fctx, buf, len, &wlen);
        break;
#endif

    default:
        return RF_ERROR;
    }
    if (ret == RF_ESUCCESS)
        *plen = wlen;
    return ret;
}

#if RF_OPT_USE_GET_FSTAT
rf_int_t rf_file_stat(const rf_char_t* pathname, rf_stat_t* buf)
{
    rf_int_t ret;
    rf_file_ctx_t fctx;

    fctx.open_flag = TRUE;
    ret = rf_file_open(&fctx, pathname, RF_O_RDONLY);
    if (ret)
        return ret;

    ret = rf_file_ioctl(&fctx, RF_FD_IOCTL_GET_FSTAT, buf);
    rf_file_close(&fctx);
    return ret;
}
#endif

#if ( RF_OPT_USE_RENAME || RF_OPT_USE_UNLINK )
rf_int_t rf_dir_not_empty(const rf_char_t* path)
{
    rf_int_t i, ret;
    rf_stat_t ffs;
    rf_dirent_t dir;
    rf_file_ctx_t fctx;

    fctx.open_flag = TRUE;
    ret = rf_file_open(&fctx, path, RF_O_RDONLY);
    if (ret)
        return ret;

    do {
        ret = rf_file_ioctl(&fctx, RF_FD_IOCTL_GET_FSTAT, &ffs);
        if (ret)
            break;
        if (ffs.st_mode & RF_ATTR_DIRECTORY) {
            dir.len = 0;
            dir.name = NULL;
            for (i = 0;; i++) {
                ret = rf_file_readdir(&fctx, "*", &dir);
                if (ret)
                    break;
            }
            /* öĿ¼ */
            if (ret != RF_ENDFILE)
                break;
            if (i > 2) {
                ret = RF_ENOTEMPTY;
                break;
            }
            ret = 0;
        }
    } while (FALSE);
    rf_file_close(&fctx);
    return ret;
}
#endif

#if RF_OPT_USE_RENAME
rf_int_t rf_file_rename(const rf_char_t* oldname, const rf_char_t* newname)
{
    rf_vol_ctx_t* old_fs, *new_fs;
    rf_stat_t source_stat;
    rf_stat_t dest_stat;
    rf_int_t dest_exists = 1;
    rf_int_t ret;

    old_fs = find_fs_ctx(oldname);
    new_fs = find_fs_ctx(oldname);
    FS_CHECK(old_fs);
    FS_CHECK(old_fs);

    if (old_fs != new_fs)
        return RF_EXDEV;

    /*
    ļ
    newnameָļڣᱻɾ
    newnameoldnameһĿ¼£൱ƶļ

    Ŀ¼
    oldnamenewnameΪĿ¼Ŀ¼
    newnameָĿ¼ΪǿĿ¼ȽnewnameĿ¼ɾ, Ŀ¼
    Ŀ¼ʱnewnameܰoldnameΪ·ǰ׺磬ܽ/usrΪ/usr/foo/testdirΪ֣/usr/fooֵ·ǰ׺ܽɾ
    */

    ret = rf_file_stat(oldname, &source_stat);
    if (ret)
        return ret;

    ret = rf_file_stat(newname, &dest_stat);
    if (ret) {
        if (ret != RF_ENOENT)
            return ret;
        dest_exists = 0;
    }

    if (dest_exists) {
        /* cannot overwrite directory with non-directory */
        if (RF_S_ISDIR(dest_stat.st_mode) && (!RF_S_ISDIR(source_stat.st_mode))) {
            return RF_EACCES;
        }

        /* cannot overwrite non-directory with directory */
        if (!RF_S_ISDIR(dest_stat.st_mode) && RF_S_ISDIR(source_stat.st_mode)) {
            return RF_EACCES;
        }

        /* ĿĿ¼  */
        ret = rf_dir_not_empty(newname);
        if (ret)
            return ret;
    }

    switch (((rf_vol_comm_t*)old_fs->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        ret = fat_rename(old_fs, oldname, newname);
        break;
#endif

    default:
        return RF_ERROR;
    }
    return ret;
}
#endif

#if RF_OPT_USE_UNLINK
rf_int_t rf_file_unlink(const rf_char_t* path)
{
    rf_int_t ret;
    rf_vol_ctx_t* vctx;

    vctx = find_fs_ctx(path);
    FS_CHECK(vctx);

    /* ֧ɾǿĿ¼ */
    ret = rf_dir_not_empty(path);
    if (ret)
        return ret;

    switch (((rf_vol_comm_t*)vctx->vol)->fs_type) {
#if RF_OPT_USE_FAT
        FS_FAT_TYPE
        ret = fat_unlink(vctx, path_get_root_dir(path));
        break;
#endif

    default:
        return RF_ERROR;
    }
    return ret;
}
#endif
