/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#ifndef __RFS_FS_H__
#define __RFS_FS_H__

#define     RFS_FILE_TYPE               1
#define     RFS_DIR_TYPE                2

#define     RF_VOL_IOCTL_STATFS         1
#define     RF_VOL_IOCTL_CHDIR          2
#define     RF_VOL_IOCTL_IOCTL_GETCWD   3

#define     RF_FD_IOCTL_GET_FSTAT       1
#define     RF_FD_IOCTL_GET_FTIME       2
#define     RF_FD_IOCTL_GET_LTIME       3
#define     RF_FD_IOCTL_GET_NAME        4
#define     RF_FD_IOCTL_GET_SHORT_NAME  5

#define     RF_FD_IOCTL_SET_ATTRIB      15
#define     RF_FD_IOCTL_SET_FTIME       16
#define     RF_FD_IOCTL_SET_LTIME       17
#define     RF_FD_IOCTL_FSYNC           18
#define     RF_FD_IOCTL_FTRUNCATE       19
#define     RF_FD_IOCTL_FDELETE         20

#define     RF_MAKE_FD_IOCTL( cmd, len )    ((len << 8) | cmd)

struct rf_vol_ctx;
struct rf_file_ctx;

struct rf_fs_ops {
    const rf_char_t*    name;
#ifdef RF_OPT_USE_MKFS
    rf_int_t (*mkfs)(const rf_char_t* devname, const rf_void_t* ctx) reentrant;
#endif
    rf_int_t (*mount)(struct rf_vol_ctx* fs, rf_mount_param_t* param) reentrant;
    rf_int_t (*unmount)(struct rf_vol_ctx* fs) reentrant;
    rf_int_t (*ioctl_vol)(struct rf_vol_ctx* fs, rf_int_t cmd, rf_void_t* args) reentrant;

    rf_int_t (*open)(struct rf_file_ctx* fd, const rf_char_t* file_name, rf_int_t flags)  reentrant;
#if RF_OPT_USE_RAWOPEN
    rf_int_t (*raw_open)(struct rf_file_ctx* fd, rf_fentry_t* fentry, rf_int_t flags)  reentrant;
#endif
    rf_int_t (*close)(struct rf_file_ctx* fd) reentrant;
    rf_int_t (*ioctl)(struct rf_file_ctx* fd, rf_int_t cmd, rf_void_t* args) reentrant;
#if RF_OPT_USE_READDIR
    rf_int_t (*readdir)(struct rf_file_ctx* fd, const rf_char_t* pattern, rf_dirent_t* dirp) reentrant;
#endif
#ifdef RF_OPT_USE_FSEEK
    rf_int_t (*lseek)(struct rf_file_ctx* fd, rf_fsize_t offset, rf_int_t origin, rf_fsize_t* now_offset) reentrant;
#endif
    rf_int_t (*read)(struct rf_file_ctx* fd, rf_void_t* buf, rf_size_t len, rf_size_t* plen) reentrant;

    rf_int_t (*write)(struct rf_file_ctx* fd, const rf_void_t* buf, rf_size_t len, rf_size_t* plen) reentrant;
#ifdef RF_OPT_USE_UNLINK
    rf_int_t (*unlink)(struct rf_vol_ctx* fs, const rf_char_t* pathname) reentrant;
#endif
    rf_int_t (*rename)(struct rf_vol_ctx* fs, const rf_char_t* oldname, const rf_char_t* newname) reentrant;
};
typedef struct rf_fs_ops rf_fs_ops_t, *rf_fs_ops_ref_t;

struct rf_vol_ctx {
    rf_int_t            open_flag: 1;
#if RF_OPT_FS_SAVE_DEV_NAME == 1
    const rf_char_t*     device_name;
#endif
    rf_void_t*           vol;
};
typedef struct rf_vol_ctx rf_vol_ctx_t, *rf_vol_ctx_ref_t;

struct rf_file_ctx {
    rf_int_t            open_flag: 1;
    rf_vol_ctx_t*        vctx;
    rf_void_t*           file;
};
typedef struct rf_file_ctx rf_file_ctx_t, *rf_file_ctx_ref_t;

#ifdef RF_OPT_PARAM_CHECK
#define FS_CHECK(ctx)   {RF_PERROR("error: invalid ctx==NULL\n", ctx == NULL, return RF_EINVAL; );\
                        RF_PERROR("error: volume didn't open\n", !ctx->open_flag, return RF_EINVAL; );\
                        RF_PERROR("error: ctx->vol==NULL\n", ctx->vol == NULL, return RF_EINVAL; );}
#else
#define FS_CHECK(ctx)
#endif

#define FS_FAT_TYPE     case RF_TYPE_FAT12:\
                        case RF_TYPE_FAT16:\
                        case RF_TYPE_FAT32:

#define FS_NTFS_TYPE    case RF_TYPE_NTFS:


#define FS_EXT_TYPE     case RF_TYPE_EXT2:\
                        case RF_TYPE_EXT3:\
                        case RF_TYPE_EXT4:


rf_vol_ctx_t* find_fs_ctx(const rf_char_t* path);

rf_int_t rf_imp_mkfs(const rf_char_t* fs_name, const rf_char_t* device_name, const rf_void_t* ctx);
rf_int_t rf_imp_mount(const rf_char_t* path, const rf_char_t* filesystemtype, rf_mount_param_t* param);
rf_int_t rf_imp_unmount(const rf_char_t* path);

rf_int_t rf_imp_statfs(const rf_char_t* path, struct rf_statfs* buf);
rf_int_t rf_imp_chdir(const rf_char_t* path);
rf_int_t rf_imp_getcwd(rf_char_t* path, rf_size_t len);

#endif
