/*!The RanFs Library
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2014-2019, www.ranfs.com
 *
*/
#include <stdarg.h>
#include "../include/rfs.h"
#include "../common/rf_common.h"
#include "rf_fs.h"
#include "rf_file.h"
#include "rf_stream.h"
#include "rf_fd.h"

rf_int_t rf_stdio_init(rf_void_t)
{
    _rf_nstream = RF_MAX_STREAM;

    /*
    * Allocate the __rf_piob array. Try for _rf_nstream entries first. If this
    * fails then reset _rf_nstream to _IOB_ENTRIES and try again. If it
    * still fails, bail out with an RTE.
    */
    if ((__rf_piob = (void**)rf_malloc(_rf_nstream * sizeof(void*))) == NULL) {
        _rf_nstream = RF_MIN_STREAM;
        if ((__rf_piob = (void**)rf_malloc(RF_MIN_STREAM * sizeof(void*))) == NULL)
            return RF_ENOMEM;
    }
    memset(__rf_piob, 0, _rf_nstream * sizeof(void*));
    return 0;
}

rf_int_t rf_stdio_exit(rf_void_t)
{
    rf_fcloseall();
    rf_free(__rf_piob);
    __rf_piob = NULL;
    _rf_nstream = 0;
    return 0;
}

static RF_FILE*   _openfile(const rf_char_t* filename, const rf_char_t* mode, RF_FILE* str)
{
    rf_int_t modeflag;
    rf_int_t streamflag = 0;
    rf_int_t commodeset = 0;
    rf_int_t scanset    = 0;
    rf_int_t whileflag;
    rf_int_t filedes;
    RF_FILE* stream;

    RF_ASSERTE("filename != NULL\n", filename != NULL);
    RF_ASSERTE("mode != NULL\n", mode != NULL);
    RF_ASSERTE("str != NULL\n", str != NULL);
    /* Parse the user's specification string as set flags in
           (1) modeflag - system call flags word
           (2) streamflag - stream handle flags word. */
    /* First mode character must be 'r', 'w', or 'a'. */

    switch (*mode) {
    case 'r':
        modeflag = RF_O_RDONLY;
        streamflag |= RF_IOREAD;
        break;
    case 'w':
        modeflag = RF_O_WRONLY | RF_O_CREAT | RF_O_TRUNC;
        streamflag |= RF_IOWRT;
        break;
    case 'a':
        modeflag = RF_O_WRONLY | RF_O_CREAT | RF_O_APPEND;
        streamflag |= RF_IOWRT | RF_IOAPPEND;
        break;
    default:
        return (NULL);
        break;
    }

    /* There can be up to three more optional mode characters:
       (1) A single '+' character,
       (2) One of 't' and 'b' and
    */
    whileflag = 1;
    while (*++mode && whileflag)
        switch (*mode) {

        case '+':
            if (modeflag & RF_O_RDWR)
                whileflag = 0;
            else {
                modeflag |= RF_O_RDWR;
                modeflag &= ~(RF_O_RDONLY | RF_O_WRONLY);
                streamflag |= RF_IORW;
                streamflag &= ~(RF_IOREAD | RF_IOWRT);
            }
            break;

        case 'b':
            if (modeflag & (RF_IOTEXT | RF_IOBINARY))
                whileflag = 0;
            else
                streamflag |= RF_IOBINARY;
            break;

        case 't':
            if (modeflag & (RF_IOTEXT | RF_IOBINARY))
                whileflag = 0;
            else
                streamflag |= RF_IOTEXT;
            break;
        default:
            whileflag = 0;
            break;
        }

    if ((filedes = rf_open(filename, modeflag)) < 0)
        return (NULL);
    stream = str;
    stream->_flag = streamflag;
    stream->_cnt = 0;
    stream->_base = stream->_ptr = NULL;
    stream->_file = filedes;
    return (stream);
}

RF_FILE* rf_fdopen(rf_int_t filedes, const rf_char_t* mode)
{
    RF_FILE* stream;
    rf_int_t whileflag, tbflag, cnflag;
    rf_int_t fileflag;
    rf_int_t _max_fd = fd_get_max_fd();

    RF_PERROR("(mode == NULL\n", mode == NULL, { rf_errno = RF_EINVAL; return NULL;});
    RF_PERROR("filedes error\n", !((filedes >= 0) && (filedes < _max_fd)), { rf_errno = RF_EBADF; return NULL; });

    /* Skip leading spaces */
    while (*mode == ' ') {
        ++mode;
    }

    /* First character must be 'r', 'w', or 'a'. */
    switch (*mode) {
    case 'r':
        fileflag = RF_IOREAD;
        break;
    case 'w':
    case 'a':
        fileflag = RF_IOWRT;
        break;
    default:
        RF_PERROR("Invalid file open mode", 1, rf_errno = RF_EINVAL; return NULL);
    }

    /* There can be up to three more optional characters:
    (1) A single '+' character,
    (2) One of 'b' and 't' and
    (3) One of 'c' and 'n'.

    Note that currently, the 't' and 'b' flags are syntax checked
    but ignored.  'c' and 'n', however, are correctly supported.
    */

    whileflag = 1;
    tbflag = cnflag = 0;
    fileflag |= 0;

    while (*++mode && whileflag)
        switch (*mode) {

        case ' ':
            /* skip spaces */
            break;

        case '+':
            if (fileflag & RF_IORW)
                whileflag = 0;
            else {
                fileflag |= RF_IORW;
                fileflag &= ~(RF_IOREAD | RF_IOWRT);
            }
            break;

        case 'b':
        case 't':
            if (tbflag)
                whileflag = 0;
            else
                tbflag = 1;
            break;

        default:
            RF_PERROR("Invalid file open mode", 1, { rf_errno = RF_EINVAL; return NULL; });
        }

    /* Skip trailing spaces */
    while (*mode == ' ') {
        ++mode;
    }
    RF_PERROR("Invalid file open mode\n", *mode != '\0', { rf_errno = RF_EINVAL; return NULL; });

    /* Find a free stream; stream is returned 'locked'. */
    if ((stream = rf_get_stream()) == NULL) {
        rf_errno = RF_EMFILE;
        return (NULL);
    }

    stream->_flag = fileflag;
    stream->_file = filedes;
    return (stream);
}

RF_FILE* rf_fopen(const rf_char_t* file, const rf_char_t* mode)
{
    RF_FILE* stream;
    RF_FILE* retval;

    RF_PERROR("file==NULL\n", file == NULL, return NULL;);
    RF_PERROR("mode==NULL\n", mode == NULL, return NULL;);

    if ((stream = rf_get_stream()) == NULL)
        return (NULL);
    retval = _openfile(file, mode, stream);
    if (retval == NULL)
        rf_free_stream(stream);
    return (retval);
}

rf_size_t rf_fread(rf_void_t* buffer, rf_size_t size, rf_size_t num, RF_FILE* stream)
{
    rf_char_t* data;                 /* point to where should be read next */
    rf_size_t total;                 /* total bytes to read */
    rf_size_t count;                 /* num bytes left to read */
    rf_size_t bufsize;               /* size of stream buffer */
    rf_size_t nbytes;                /* how much to read now */
    rf_size_t nread;                 /* how much we did read */
    rf_int_t c;                      /* a temp char */

    RF_PERROR("buffer==NULL\n", buffer == NULL, return -1;);
    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);

    /* initialize local vars */
    data = buffer;
    if ((count = total = size * num) == 0)
        return 0;

    if (rf_anybuf(stream))
        /* already has buffer, use its size */
        bufsize = stream->_bufsiz;
    else
        bufsize = RF_SMALL_BUFSIZ;

    /* here is the main loop -- we go through here until we're done */
    while (count != 0) {
        /* if the buffer exists and has characters, copy them to user
           buffer */
        if (rf_anybuf(stream) && stream->_cnt != 0) {
            /* how much do we want? */
            nbytes = (count < (unsigned)stream->_cnt) ? count : stream->_cnt;
            memcpy(data, stream->_ptr, nbytes);

            /* update stream and amt of data read */
            count -= nbytes;
            stream->_cnt -= nbytes;
            stream->_ptr += nbytes;
            data += nbytes;
        } else if (count >= bufsize) {
            /* If we have more than bufsize chars to read, get data
               by calling read with an integral number of bufsiz
               blocks.  Note that if the stream is text mode, read
               will return less chars than we ordered. */

            /* calc chars to read -- (count/bufsize) * bufsize */
            nbytes = (bufsize ? (count - count % bufsize) : count);
            nread = rf_read(rf_fileno(stream), data, nbytes);
            if (nread == 0) {
                /* end of file -- out of here */
                stream->_flag |= RF_IOEOF;
                return (total - count) / size;
            } else if (nread == (unsigned) - 1) {
                /* error -- out of here */
                stream->_flag |= RF_IOERR;
                return (total - count) / size;
            }

            /* update count and data to reflect read */
            count -= nread;
            data += nread;
        } else {
            /* less than bufsize chars to read, so call _filbuf to fill buffer */
            if ((c = rf_filbuf(stream)) == RF_EOF) {
                /* error or eof, stream flags set by _filbuf */
                return (total - count) / size;
            }

            /* _filbuf returned a char -- store it */
            *data++ = (char) c;
            --count;

            /* update buffer size */
            bufsize = stream->_bufsiz;
        }
    }

    /* we finished successfully, so just return num */
    return num;
}

rf_int_t rf_flsbuf(rf_int_t ch, RF_FILE* stream)
{
    rf_ssize_t charcount;
    rf_ssize_t written;
    rf_int_t fh;

    RF_ASSERTE("stream=NULL\n", stream != NULL);

    fh = rf_fileno(stream);
    if (!(stream->_flag & (RF_IOWRT | RF_IORW))) {
        rf_errno = RF_EBADF;
        stream->_flag |= RF_IOERR;
        return (RF_EOF);
    } else if ((stream->_flag & RF_IOSTRG)) {
        rf_errno = RF_EACCES;
        stream->_flag |= RF_IOERR;
        return (RF_EOF);
    }

    /* Check that _IOREAD is not set or, if it is, then so is _IOEOF. Note
    that _IOREAD and IOEOF both being set implies switching from read to
    write at end-of-file, which is allowed by ANSI. Note that resetting
    the _cnt and _ptr fields amounts to doing an fflush() on the stream
    in this case. Note also that the _cnt field has to be reset to 0 for
    the error path as well (i.e., _IOREAD set but _IOEOF not set) as
    well as the non-error path. */

    if (stream->_flag & RF_IOREAD) {
        stream->_cnt = 0;
        if (stream->_flag & RF_IOEOF) {
            stream->_ptr = stream->_base;
            stream->_flag &= ~RF_IOREAD;
        } else {
            stream->_flag |= RF_IOERR;
            return (RF_EOF);
        }
    }

    stream->_flag |= RF_IOWRT;
    stream->_flag &= ~RF_IOEOF;
    written = charcount = stream->_cnt = 0;

    /* Get a buffer for this stream, if necessary. */
    if (!rf_anybuf(stream)) {

        /* Do NOT get a buffer if (1) stream is stdout/stderr, and
        (2) stream is NOT a tty.
        [If stdout/stderr is a tty, we do NOT set up single char
        buffering. This is so that later temporary buffering will
        not be thwarted by the _IONBF bit being set (see
        _stbuf/_ftbuf usage).]
        if (!(((stream == stdout) || (stream == stderr)) && (_isatty(fh))))

        */
        rf_getbuf(stream);

    } /* end !anybuf() */

    /* If big buffer is assigned to stream... */
    if (rf_bigbuf(stream)) {

        RF_ASSERTE("inconsistent IOB fields\n", stream->_ptr - stream->_base >= 0);

        charcount = (int)(stream->_ptr - stream->_base);
        stream->_ptr = stream->_base + sizeof(rf_char_t);
        stream->_cnt = stream->_bufsiz - (int)sizeof(rf_char_t);

        if (charcount > 0)
            written = rf_write(fh, stream->_base, charcount);
        else if (stream->_flag & RF_IOAPPEND) {
            if (rf_lseek(fh, 0L, RF_SEEK_END) == -1) {
                stream->_flag |= RF_IOERR;
                return (RF_EOF);
            }
        }

        *stream->_base = (rf_u8_t)ch;
    }

    /* Perform single character output (either _IONBF or no buffering) */
    else {
        charcount = sizeof(rf_char_t);
        written = rf_write(fh, &ch, charcount);
    }

    /* See if the _write() was successful. */
    if (written != charcount) {
        stream->_flag |= RF_IOERR;
        return (RF_EOF);
    }

    return (ch & 0xff);
}

rf_size_t rf_fwrite(const rf_void_t* buffer, rf_size_t size, rf_size_t num, RF_FILE* stream)
{
    const char* data;                 /* point to where data comes from next */
    rf_ssize_t total;                 /* total bytes to write */
    rf_ssize_t count;                 /* num bytes left to write */
    rf_ssize_t bufsize;               /* size of stream buffer */
    rf_ssize_t nbytes;                /* number of bytes to write now */
    rf_ssize_t nwritten;              /* number of bytes written */
    rf_int_t c;                       /* a temp char */

    RF_PERROR("buffer==NULL\n", buffer == NULL, return -1;);
    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);

    /* initialize local vars */
    data = buffer;
    count = total = size * num;
    if (0 == count)
        return 0;

    if (rf_anybuf(stream))
        /* already has buffer, use its size */
        bufsize = stream->_bufsiz;
    else
        bufsize = RF_SMALL_BUFSIZ;

    /* here is the main loop -- we go through here until we're done */
    while (count != 0) {
        /* if the buffer is big and has room, copy data to buffer */
        if (rf_bigbuf(stream) && stream->_cnt != 0) {
            /* how much do we want? */
            nbytes = (count < (unsigned)stream->_cnt) ? count : stream->_cnt;
            memcpy(stream->_ptr, data, nbytes);

            /* update stream and amt of data written */
            count -= nbytes;
            stream->_cnt -= nbytes;
            stream->_ptr += nbytes;
            data += nbytes;
        } else if (count >= bufsize) {
            /* If we have more than bufsize chars to write, write
               data by calling write with an integral number of
               bufsiz blocks.  If we reach here and we have a big
               buffer, it must be full so _flush it. */
            if (rf_bigbuf(stream)) {
                if (rf_fflush(stream)) {
                    /* error, stream flags set -- we're out
                       of here */
                    return (total - count) / size;
                }
            }

            /* calc chars to read -- (count/bufsize) * bufsize */
            nbytes = (bufsize ? (count - count % bufsize) : count);
            nwritten = rf_write(rf_fileno(stream), data, nbytes);
            if (nwritten == (unsigned)RF_EOF) {
                /* error -- out of here */
                stream->_flag |= RF_IOERR;
                return (total - count) / size;
            }

            /* update count and data to reflect write */
            count -= nwritten;
            data += nwritten;

            if (nwritten < nbytes) {
                /* error -- out of here */
                stream->_flag |= RF_IOERR;
                return (total - count) / size;
            }
        } else {
            /* buffer full and not enough chars to do direct write,
               so do a _flsbuf. */
            c = *data;  /* _flsbuf write one char, this is it */
            if (rf_flsbuf(c, stream) == EOF) {
                /* error or eof, stream flags set by _flsbuf */
                return (total - count) / size;
            }

            /* _flsbuf wrote a char -- update count */
            ++data;
            --count;

            /* update buffer size */
            bufsize = stream->_bufsiz > 0 ? stream->_bufsiz : 1;
        }
    }
    /* we finished successfully, so just return num */
    return num;
}

rf_int_t rf_fputc(rf_int_t ch, RF_FILE* stream)
{
    rf_int_t retval;

    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);

    retval = rf_putc_lk(ch, stream);
    return (retval);
}

rf_int_t rf_fputs(const rf_char_t* string, RF_FILE* stream)
{
    rf_ssize_t length;
    rf_int_t ret;

    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);
    RF_PERROR("string==NULL\n", string == NULL, return -1;);

    length = strlen(string);
    ret = (rf_int_t)rf_fwrite(string, 1, length, stream);
    return (ret == length ? 0 : RF_EOF);
}

rf_int_t rf_output(RF_FILE* _File, const rf_char_t* _Format, va_list _ArgList)
{
    return 0;
}

rf_int_t rf_fprintf(RF_FILE* stream, const char* format, ...)
{
    rf_char_t buf[RF_MAX_FPRINTF_BUFSIZ];
    rf_ssize_t ret = -1, count;

    va_list(arglist);
    va_start(arglist, format);
    count = vsnprintf(buf, sizeof(buf), format, arglist);
    va_end(arglist);
    if (count > 0) {
        ret = rf_fwrite(buf, (rf_size_t)1, (rf_size_t)count, stream);
        if (ret < count)
            ret = -1;
    }
    return (rf_int_t)ret;
}

rf_int_t rf_fflush(RF_FILE* stream)
{
    rf_int_t rc = 0; /* assume good return */
    rf_ssize_t nchar;

    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);
    /* Init pointer to stream */
    if ((stream->_flag & (RF_IOREAD | RF_IOWRT)) == RF_IOWRT && rf_bigbuf(stream)
            && (nchar = stream->_ptr - stream->_base) > 0) {
        if (rf_write(rf_fileno(stream), stream->_base, nchar) == nchar) {
            /* if this is a read/write file, clear _IOWRT so that
             * next operation can be a read
             */
            if (RF_IORW & stream->_flag)
                stream->_flag &= ~RF_IOWRT;
        } else {
            stream->_flag |= RF_IOERR;
            rc = RF_EOF;
        }
    }

    stream->_ptr = stream->_base;
    stream->_cnt = 0;
    return (rc);
}

rf_off_t rf_ftell(RF_FILE* stream)
{
    rf_off_t offset;
    rf_off_t filepos;
    rf_int_t fd;
    rf_off_t rdcnt;

    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);

    /* Init stream pointer and file descriptor */
    fd = rf_fileno(stream);
    if (stream->_cnt < 0)
        stream->_cnt = 0;

    if ((filepos = (rf_off_t)rf_lseek(fd, 0, RF_SEEK_CUR)) < 0)
        return (-1);

    if (!rf_bigbuf(stream))            /* _IONBF or no buffering designated */
        return (filepos - stream->_cnt);

    offset = stream->_ptr - stream->_base;
    if (stream->_flag & (RF_IOWRT | RF_IOREAD)) {
#if 0
        if (rf_osfile(fd) & RF_FTEXT)
            for (p = stream->_base; p < stream->_ptr; p++)
                if (*p == '\n')  /* adjust for '\r' */
                    offset++;
#endif
    } else if (!(stream->_flag & RF_IORW)) {
        rf_errno = RF_EINVAL;
        return (-1);
    }

    if (filepos == 0)
        return ((rf_u64_t)offset);

    if (stream->_flag & RF_IOREAD)    /* go to preceding sector */

        if (stream->_cnt == 0)  /* filepos holds correct location */
            offset = 0;
        else {

            /* Subtract out the number of unread bytes left in the
               buffer. [We can't simply use _iob[]._bufsiz because
               the last read may have hit EOF and, thus, the buffer
               was not completely filled.] */
            rdcnt = stream->_cnt + (stream->_ptr - stream->_base);
#if 0
            /* If text mode, adjust for the cr/lf substitution. If
               binary mode, we're outta here. */
            if (_osfile(fd) & FTEXT) {
                /* (1) If we're not at eof, simply copy _bufsiz
                   onto rdcnt to get the # of untranslated
                   chars read. (2) If we're at eof, we must
                   look through the buffer expanding the '\n'
                   chars one at a time. */

                /* [NOTE: Performance issue -- it is faster to
                   do the two _lseek() calls than to blindly go
                   through and expand the '\n' chars regardless
                   of whether we're at eof or not.] */

                if (_lseeki64(fd, 0, SEEK_END) == filepos) {

                    max = stream->_base + rdcnt;
                    for (p = stream->_base; p < max; p++)
                        if (*p == '\n')
                            /* adjust for '\r' */
                            rdcnt++;

                    /* If last byte was ^Z, the lowio read
                       didn't tell us about it.  Check flag
                       and bump count, if necessary. */

                    if (stream->_flag & _IOCTRLZ)
                        ++rdcnt;
                }

                else {

                    _lseeki64(fd, filepos, SEEK_SET);

                    /* We want to set rdcnt to the number
                       of bytes originally read into the
                       stream buffer (before crlf->lf
                       translation). In most cases, this
                       will just be _bufsiz. However, the
                       buffer size may have been changed,
                       due to fseek optimization, at the
                       END of the last _filbuf call. */

                    if ((rdcnt <= _SMALL_BUFSIZ) &&
                            (stream->_flag & _IOMYBUF) &&
                            !(stream->_flag & _IOSETVBUF)) {
                        /* The translated contents of
                           the buffer is small and we
                           are not at eof. The buffer
                           size must have been set to
                           _SMALL_BUFSIZ during the
                           last _filbuf call. */

                        rdcnt = _SMALL_BUFSIZ;
                    } else
                        rdcnt = stream->_bufsiz;


                    /* If first byte in untranslated buffer
                       was a '\n', assume it was preceeded
                       by a '\r' which was discarded by the
                       previous read operation and count
                       the '\n'. */
                    if (_osfile(fd) & FCRLF)
                        ++rdcnt;
                }

            } /* end if FTEXT */
#endif

            filepos -= (rf_u64_t)rdcnt;

        } /* end else stream->_cnt != 0 */

    return (filepos + (rf_u64_t)offset);
}

rf_int_t rf_fseek(RF_FILE* stream, rf_off_t offset, rf_int_t whence)
{
    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);
    if (!rf_inuse(stream) || ((whence != RF_SEEK_SET) && (whence != RF_SEEK_CUR) &&
                              (whence != RF_SEEK_END))) {
        rf_errno = RF_EINVAL;
        return (-1);
    }

    /* Clear EOF flag */
    stream->_flag &= ~RF_IOEOF;

    /* If seeking relative to current location, then convert to
       a seek relative to beginning of file.  This accounts for
       buffering, etc. by letting fseek() tell us where we are. */
    if (whence == RF_SEEK_CUR) {
        offset += rf_ftell(stream);
        whence = RF_SEEK_SET;
    }

    /* Flush buffer as necessary */
    rf_fflush(stream);

    /* If file opened for read/write, clear flags since we don't know
       what the user is going to do next. If the file was opened for
       read access only, decrease _bufsiz so that the next _filbuf
       won't cost quite so much */

    if (stream->_flag & RF_IORW)
        stream->_flag &= ~(RF_IOWRT | RF_IOREAD);
#if !defined (_M_MPPC) && !defined (_M_M68K)
    else if ((stream->_flag & RF_IOREAD) && (stream->_flag & RF_IOMYBUF) &&
             !(stream->_flag & RF_IOSETVBUF))
        stream->_bufsiz = RF_SMALL_BUFSIZ;
#endif  /* !defined (_M_MPPC) && !defined (_M_M68K) */

    /* Seek to the desired locale and return. */
    return (rf_lseek(rf_fileno(stream), (rf_off_t)offset, whence) == -1L ? -1 : 0);
}

rf_int_t rf_rewind(RF_FILE* stream)
{
    rf_int_t fd;
    rf_int_t ret;

    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);
    /* Init stream pointer */
    fd = rf_fileno(stream);
    /* Flush the stream */
    ret = rf_fflush(stream);
    if (ret)
        return ret;

    /* Clear errors */
    stream->_flag &= ~(RF_IOERR | RF_IOEOF);
#if 0
#ifdef _WIN32
    _osfile_safe(fd) &= ~(FEOFLAG);
#else  /* _WIN32 */
    _osfile[fd] &= ~(RF_FEOFLAG);
#endif  /* _WIN32 */
#endif

    /* Set flags */
    /* [note: _flush set _cnt=0 and _ptr=_base] */
    if (stream->_flag & RF_IORW)
        stream->_flag &= ~(RF_IOREAD | RF_IOWRT);

    /* Position to beginning of file */
    if (rf_lseek(fd, 0L, 0) < 0)
        return -1;
    return 0;
}

void rf_clearerr(RF_FILE* stream)
{
    RF_PERROR("stream==NULL\n", stream == NULL, return;);
    /* Clear stdio level flags */
    stream->_flag &= ~(RF_IOERR | RF_IOEOF);

    /* Clear lowio level flags */
#ifdef _WIN32
    /*_osfile_safe(_fileno(stream)) &= ~(FEOFLAG); */
#else  /* _WIN32 */
    /*_osfile[_fileno(stream)] &= ~(FEOFLAG); */
#endif  /* _WIN32 */
}

rf_int_t rf_fclose(RF_FILE* stream)
{
    rf_int_t result = RF_EOF;

    RF_PERROR("stream==NULL\n", stream == NULL, return -1;);
    if (stream->_flag & RF_IOSTRG) {
        stream->_flag = 0;
        return (RF_EOF);
    }

    if (rf_inuse(stream)) {

        /* Stream is in use:
               (1) flush stream
               (2) free the buffer
               (3) close the file
               (4) delete the file if temporary
        */

        result = rf_fflush(stream);
        rf_free_stream(stream);
        if (rf_close(rf_fileno(stream)) < 0)
            result = RF_EOF;
    }
    stream->_flag = 0;
    return (result);
}

rf_int_t rf_fcloseall(rf_void_t)
{
    rf_int_t count = 0;
    rf_int_t i;

    for (i = 0; i < _rf_nstream; i++) {

        if (__rf_piob[i] != NULL) {
            /*
            * if the stream is in use, close it
            */
            if (rf_inuse((RF_FILE*)__rf_piob[i]) && (rf_fclose(__rf_piob[i]) != RF_EOF))
                count++;
            rf_free(__rf_piob[i]);
            __rf_piob[i] = NULL;
        }
    }
    return (count);
}
